# -*- coding: utf-8 -*-

"""Utilities for :mod:`pybel.struct`."""

from ..constants import (
    GRAPH_ANNOTATION_LIST,
    GRAPH_ANNOTATION_PATTERN,
    GRAPH_ANNOTATION_URL,
    GRAPH_NAMESPACE_PATTERN,
    GRAPH_NAMESPACE_URL,
)

__all__ = [
    "update_metadata",
]


def update_metadata(source, target) -> None:
    """Update the namespace and annotation metadata in the target graph.

    :param pybel.BELGraph source:
    :param pybel.BELGraph target:
    """
    target.namespace_url.update(source.graph.get(GRAPH_NAMESPACE_URL, {}))
    target.namespace_pattern.update(source.graph.get(GRAPH_NAMESPACE_PATTERN, {}))

    target.annotation_url.update(source.graph.get(GRAPH_ANNOTATION_URL, {}))
    target.annotation_pattern.update(source.graph.get(GRAPH_ANNOTATION_PATTERN, {}))
    for keyword, values in source.graph.get(GRAPH_ANNOTATION_LIST, {}).items():
        if keyword not in target.annotation_list:
            target.annotation_list[keyword] = values
        else:
            target.annotation_list[keyword].update(values)



# -*- coding: utf-8 -*-

"""Internal DSL functions for edges."""

import warnings
from typing import Dict, Optional, Union

from ..constants import (
    ACTIVITY,
    CELL_SURFACE,
    DEGRADATION,
    EFFECT,
    EXTRACELLULAR,
    FROM_LOC,
    INTRACELLULAR,
    LOCATION,
    MODIFIER,
    NAME,
    NAMESPACE,
    TO_LOC,
    TRANSLOCATION,
)
from ..language import Entity, activity_mapping, compartment_mapping

__all__ = [
    "activity",
    "degradation",
    "translocation",
    "secretion",
    "cell_surface_expression",
    "location",
]

ModifierDict = Dict
LocationDict = Dict


def _modifier_helper(
    modifier: str,
    location: Optional[LocationDict] = None,
) -> ModifierDict:
    """Make a modifier dictionary.

    :param modifier: The name of the modifier
    :param location: An entity from :func:`pybel.dsl.entity`
    """
    rv = {
        MODIFIER: modifier,
    }

    if location:
        rv[LOCATION] = location

    return rv


def activity(
    name: Optional[str] = None,
    namespace: Optional[str] = None,
    identifier: Optional[str] = None,
    location: Optional[LocationDict] = None,
) -> ModifierDict:
    """Make a subject/object modifier dictionary.

    :param name: The name of the activity. If no namespace given, uses BEL default namespace
    :param namespace: The namespace of the activity
    :param identifier: The identifier of the name in the database
    :param location: An entity from :func:`pybel.dsl.entity` representing the location of the node
    """
    rv = _modifier_helper(ACTIVITY, location=location)

    if name and not namespace:
        rv[EFFECT] = activity_mapping[name]
    elif not name and not namespace and not identifier:
        rv[EFFECT] = activity_mapping["act"]
    else:
        rv[EFFECT] = Entity(
            namespace=namespace,
            name=name,
            identifier=identifier,
        )
    return rv


def degradation(location: Optional[LocationDict] = None) -> ModifierDict:
    """Make a degradation dictionary.

    :param location: An entity from :func:`pybel.dsl.entity` representing the location of the node
    """
    return _modifier_helper(DEGRADATION, location=location)


def translocation(
    from_loc: Union[str, Entity],
    to_loc: Union[str, Entity],
) -> ModifierDict:
    """Make a translocation dictionary.

    :param dict from_loc: An entity dictionary from :func:`pybel.dsl.entity`
    :param dict to_loc: An entity dictionary from :func:`pybel.dsl.entity`
    :rtype: dict
    """
    rv = _modifier_helper(TRANSLOCATION)
    if isinstance(from_loc, str):
        from_loc = compartment_mapping[from_loc]
    if not isinstance(from_loc, Entity):
        raise TypeError
    if isinstance(to_loc, str):
        to_loc = compartment_mapping[to_loc]
    if not isinstance(to_loc, Entity):
        raise TypeError

    rv[EFFECT] = {
        FROM_LOC: from_loc,
        TO_LOC: to_loc,
    }
    return rv


def secretion() -> ModifierDict:
    """Make a secretion translocation dictionary.

    This is a convenient wrapper representing the :func:`translocation` from the intracellular location to the
    extracellular space.
    """
    return translocation(INTRACELLULAR, EXTRACELLULAR)


def cell_surface_expression() -> ModifierDict:
    """Make a cellular surface expression translocation dictionary.

    This is a convenient wrapper representing the :func:`translocation` from the intracellular location to the cell
    surface.
    """
    return translocation(INTRACELLULAR, CELL_SURFACE)


def location(identifier: Entity) -> LocationDict:
    """Make a location object modifier dictionary.

    :param identifier: A namespace/name/identifier pair

    Usage:

    X increases the abundance of Y in the cytoplasm

    .. code-block:: python

        from pybel import BELGraph
        from pybel.dsl import protein, location

        graph = BELGraph()

        source = protein('HGNC', 'IRAK1')
        target = protein('HGNC', 'IRF7, variants=[
            pmod('Ph', 'Ser', 477), pmod('Ph', 'Ser', 479),
        ])
        graph.add_increases(
            source,
            target,
            citation=...,
            evidence=...,
            target_modifier=location(entity(namespace='GO', name='cytosol', identifier='GO:0005829')),
        )

    X increases the kinase activity of Y in the cytoplasm. In this case, the :func:`activity` function takes a location as
    an optional argument.

    .. code-block:: python

        from pybel import BELGraph
        from pybel.dsl import protein, location

        graph = BELGraph()

        source = ...
        target = ...
        graph.add_increases(
            source,
            target,
            citation=...,
            evidence=...,
            target_modifier=activity('kin', location=entity(namespace='GO', name='cytosol', identifier='GO:0005829')),
        )
    """
    return {
        LOCATION: identifier,
    }



# -*- coding: utf-8 -*-

"""Tests for metadata transforations."""

import unittest

from pybel import BELGraph
from pybel.constants import (
    ANNOTATIONS,
    CITATION,
    CITATION_AUTHORS,
    CITATION_DATE,
    CITATION_TYPE_PUBMED,
    IDENTIFIER,
    NAMESPACE,
)
from pybel.dsl import protein
from pybel.examples import sialic_acid_graph
from pybel.struct.mutation import (
    add_annotation_value,
    remove_annotation_value,
    remove_extra_citation_metadata,
    strip_annotations,
)
from pybel.testing.utils import n


class TestMetadata(unittest.TestCase):
    """Test metadata transformations."""

    def test_strip_annotations(self):
        """Test the strip_annotation function."""
        x = protein(namespace="HGNC", name="X")
        y = protein(namespace="HGNC", name="X")

        graph = BELGraph()
        graph.annotation_list["A"] = set("ABC")
        key = graph.add_increases(
            x,
            y,
            citation="123456",
            evidence="Fake",
            annotations={"A": {"B": True}},
        )

        self.assertIn(ANNOTATIONS, graph[x][y][key])

        strip_annotations(graph)
        self.assertNotIn(ANNOTATIONS, graph[x][y][key])

    def test_add_and_remove_annotation(self):
        """Test adding and removing annotations.

        See: :func:`pybel.struct.mutation.add_annotation_value` and
        :func:`pybel.struct.mutation.remove_annotation_value` functions.
        """
        graph = sialic_acid_graph.copy()
        annotation = "test-annotation"
        value = "test-value"
        url = n()

        graph.annotation_url[annotation] = url

        add_annotation_value(graph, annotation, value)

        for u, v, d in graph.edges(data=True):
            annotations = d.get(ANNOTATIONS)

            if annotations is None:
                continue

            self.assertIn(annotation, annotations)
            self.assertIn(value, annotations[annotation])

        remove_annotation_value(graph, annotation, value)

        for u, v, d in graph.edges(data=True):
            annotations = d.get(ANNOTATIONS)

            if annotations is None:
                continue

            annotation_values = annotations.get(annotation)

            if annotation_values is None:
                continue

            self.assertNotIn(value, annotation_values)

    def test_remove_citation_metadata(self):
        """Test removing citation metadata from a graph."""
        x = protein(namespace="HGNC", name="X")
        y = protein(namespace="HGNC", name="X")

        graph = BELGraph()
        graph.annotation_list["A"] = set("ABC")

        k0 = graph.add_part_of(x, y)
        k1 = graph.add_increases(
            x,
            y,
            citation="123456",
            evidence="Fake",
            annotations={"A": {"B": True}},
        )
        k2 = graph.add_increases(
            x,
            y,
            citation={
                NAMESPACE: CITATION_TYPE_PUBMED,
                IDENTIFIER: "12345678",
                CITATION_DATE: "2018-12-10",
            },
            evidence="Fake",
            annotations={"A": {"B": True}},
        )

        remove_extra_citation_metadata(graph)

        self.assertNotIn(CITATION, graph[x][y][k0])

        for k in k1, k2:
            self.assertIn(CITATION, graph[x][y][k])
            self.assertIn(NAMESPACE, graph[x][y][k][CITATION])
            self.assertIn(IDENTIFIER, graph[x][y][k][CITATION])
            self.assertNotIn(CITATION_DATE, graph[x][y][k][CITATION])
            self.assertNotIn(CITATION_AUTHORS, graph[x][y][k][CITATION])



# -*- coding: utf-8 -*-

"""Test cases for PyBEL testing."""

import logging
import os
import tempfile
import unittest

import pystow

from ..manager import Manager

__all__ = [
    "TEST_CONNECTION",
    "TemporaryCacheMixin",
    "TemporaryCacheClsMixin",
    "FleetingTemporaryCacheMixin",
]

logger = logging.getLogger(__name__)

TEST_CONNECTION = pystow.get_config("pybel", "test_connection")


class TemporaryCacheMixin(unittest.TestCase):
    """A test case that has a connection and a manager that is created for each test function."""

    def setUp(self):
        """Set up the test function with a connection and manager."""
        if TEST_CONNECTION:
            self.connection = TEST_CONNECTION
        else:
            self.fd, self.path = tempfile.mkstemp()
            self.connection = "sqlite:///" + self.path
            logger.info("Test generated connection string %s", self.connection)

        self.manager = Manager(connection=self.connection, autoflush=True)
        self.manager.create_all()

    def tearDown(self):
        """Tear down the test function by closing the session and removing the database."""
        self.manager.session.close()

        if not TEST_CONNECTION:
            os.close(self.fd)
            os.remove(self.path)
        else:
            self.manager.drop_all()


class TemporaryCacheClsMixin(unittest.TestCase):
    """A test case that has a connection and a manager that is created for each test class."""

    fd, path, manager = None, None, None

    @classmethod
    def setUpClass(cls):
        """Set up the test class with a connection and manager."""
        if TEST_CONNECTION:
            cls.connection = TEST_CONNECTION
        else:
            cls.fd, cls.path = tempfile.mkstemp()
            cls.connection = "sqlite:///" + cls.path
            logger.info("Test generated connection string %s", cls.connection)

        cls.manager = Manager(connection=cls.connection, autoflush=True)
        cls.manager.create_all()

    @classmethod
    def tearDownClass(cls):
        """Tear down the test class by closing the session and removing the database."""
        cls.manager.session.close()

        if not TEST_CONNECTION:
            os.close(cls.fd)
            os.remove(cls.path)
        else:
            cls.manager.drop_all()


class FleetingTemporaryCacheMixin(TemporaryCacheClsMixin):
    """A test case that clears the database before each function."""

    def setUp(self):
        """Set up the function by clearing the database."""
        super(FleetingTemporaryCacheMixin, self).setUp()

        self.manager.drop_networks()
        self.manager.drop_edges()
        self.manager.drop_nodes()
        self.manager.drop_namespaces()



# -*- coding: utf-8 -*-

"""Citation utilities for the database manager."""

import logging
import re
from datetime import date, datetime
from functools import lru_cache
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Tuple, Union

import ratelimit
import requests
from more_itertools import chunked
from sqlalchemy import and_
from tqdm.autonotebook import tqdm

from . import models
from .cache_manager import Manager
from ..constants import CITATION
from ..struct.filters import filter_edges
from ..struct.filters.edge_predicates import CITATION_PREDICATES
from ..struct.graph import BELGraph
from ..struct.summary.provenance import get_citation_identifiers

__all__ = [
    "enrich_pubmed_citations",
    "enrich_pmc_citations",
]

logger = logging.getLogger(__name__)

EUTILS_URL_FMT = "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?db=pubmed&retmode=json&id={}"

re1 = re.compile(r"^[12][0-9]{3} [a-zA-Z]{3} \d{1,2}$")
re2 = re.compile(r"^[12][0-9]{3} [a-zA-Z]{3}$")
re3 = re.compile(r"^[12][0-9]{3}$")
re4 = re.compile(r"^[12][0-9]{3} [a-zA-Z]{3}-[a-zA-Z]{3}$")
re5 = re.compile(r"^([12][0-9]{3}) (Spring|Fall|Winter|Summer)$")
re6 = re.compile(r"^[12][0-9]{3} [a-zA-Z]{3} \d{1,2}-(\d{1,2})$")
re7 = re.compile(r"^[12][0-9]{3} [a-zA-Z]{3} \d{1,2}-([a-zA-Z]{3} \d{1,2})$")

# TODO "Winter 2016" probably with re.compile(r'^(Spring|Fall|Winter|Summer) ([12][0-9]{3})$')
# TODO "YYYY Oct - Dec" update re4 to allow spaces before and after the dash

season_map = {"Spring": "03", "Summer": "06", "Fall": "09", "Winter": "12"}


def sanitize_date(publication_date: str) -> str:
    """Sanitize lots of different date strings into ISO-8601."""
    if re1.search(publication_date):
        return datetime.strptime(publication_date, "%Y %b %d").strftime("%Y-%m-%d")

    if re2.search(publication_date):
        return datetime.strptime(publication_date, "%Y %b").strftime("%Y-%m-01")

    if re3.search(publication_date):
        return publication_date + "-01-01"

    if re4.search(publication_date):
        return datetime.strptime(publication_date[:-4], "%Y %b").strftime("%Y-%m-01")

    s = re5.search(publication_date)

    if s:
        year, season = s.groups()
        return "{}-{}-01".format(year, season_map[season])

    s = re6.search(publication_date)

    if s:
        return datetime.strptime(publication_date, "%Y %b %d-{}".format(s.groups()[0])).strftime("%Y-%m-%d")

    s = re7.search(publication_date)

    if s:
        return datetime.strptime(publication_date, "%Y %b %d-{}".format(s.groups()[0])).strftime("%Y-%m-%d")


def clean_pubmed_identifiers(identifiers: Iterable[str]) -> List[str]:
    """Clean a list of identifiers with string strips, deduplicates, and sorting."""
    _identifiers = (str(identifier).strip() for identifier in identifiers if identifier)
    return sorted({i for i in _identifiers if i})


@ratelimit.limits(calls=3, period=1)
def get_pubmed_citation_response(pubmed_identifiers: Iterable[str]):
    """Get the response from PubMed E-Utils for a given list of PubMed identifiers.

    Rate limit of 3 requests per second is from:
    https://ncbiinsights.ncbi.nlm.nih.gov/2018/08/14/release-plan-for-e-utility-api-keys/

    :param pubmed_identifiers:
    :rtype: dict
    """
    pubmed_identifiers = list(pubmed_identifiers)
    url = EUTILS_URL_FMT.format(
        ",".join(pubmed_identifier for pubmed_identifier in pubmed_identifiers if pubmed_identifier),
    )
    response = requests.get(url)
    return response.json()


def enrich_citation_model(manager: Manager, citation: models.Citation, p: Mapping[str, Any]) -> bool:
    """Enrich a citation model with the information from PubMed.

    :param manager: A database manager
    :param citation: A citation model
    :param p: The dictionary from PubMed E-Utils corresponding to d["result"][pmid]
    """
    if "error" in p:
        logger.warning("Error downloading PubMed")
        return False

    citation.title = p["title"]
    citation.journal = p["fulljournalname"]
    citation.volume = p["volume"]
    citation.issue = p["issue"]
    citation.pages = p["pages"]
    citation.first = manager.get_or_create_author(p["sortfirstauthor"])
    citation.last = manager.get_or_create_author(p["lastauthor"])
    pubtypes = p["pubtype"]
    if pubtypes:
        citation.article_type = pubtypes[0]

    if "authors" in p:
        for author in p["authors"]:
            author_model = manager.get_or_create_author(author["name"])
            if author_model not in citation.authors:
                citation.authors.append(author_model)

    publication_date = p["pubdate"]
    try:
        sanitized_publication_date = sanitize_date(publication_date)
    except ValueError:
        logger.warning(
            "could not parse publication date %s for pubmed:%s",
            publication_date,
            citation.db_id,
        )
        sanitized_publication_date = None

    if sanitized_publication_date:
        citation.date = datetime.strptime(sanitized_publication_date, "%Y-%m-%d")
    else:
        logger.info("result had date with strange format: %s", publication_date)

    return True


def get_citations_by_pmids(
    manager: Manager,
    pmids: Iterable[Union[str, int]],
    *,
    group_size: Optional[int] = None,
    offline: bool = False,
) -> Tuple[Dict[str, Dict], Set[str]]:
    return _get_citations_by_identifiers(
        manager=manager,
        identifiers=pmids,
        group_size=group_size,
        offline=offline,
        prefix="pubmed",
    )


def _get_citations_by_identifiers(
    manager: Manager,
    identifiers: Iterable[Union[str, int]],
    *,
    group_size: Optional[int] = None,
    offline: bool = False,
    prefix: Optional[str] = None,
) -> Tuple[Dict[str, Dict], Set[str]]:
    """Get citation information for the given list of PubMed identifiers using the NCBI's eUtils service.

    :type manager: pybel.Manager
    :param identifiers: an iterable of PubMed identifiers
    :param group_size: The number of PubMed identifiers to query at a time. Defaults to 200 identifiers.
    :return: A dictionary of {identifier: data dictionary} or a pair of this dictionary and a set ot erroneous
             identifiers.
    """
    if prefix is None:
        prefix = "pubmed"

    helper = _HELPERS.get(prefix)
    if helper is None:
        raise ValueError(f"can not work on prefix: {prefix}")

    group_size = group_size if group_size is not None else 200

    identifiers = clean_pubmed_identifiers(identifiers)
    logger.info("ensuring %d %s identifiers", len(identifiers), prefix)

    enriched_models = {}
    unenriched_models = {}

    id_to_model = {
        citation_model.db_id: citation_model
        for citation_model in _get_citation_models(identifiers, prefix=prefix, manager=manager)
    }
    logger.info(
        "%d of %d %s identifiers are already cached",
        len(id_to_model),
        len(identifiers),
        prefix,
    )
    for identifier in tqdm(identifiers, desc=f"creating {prefix} models"):
        model = id_to_model.get(identifier)
        if model is None:
            model = id_to_model[identifier] = manager.get_or_create_citation(identifier=identifier, namespace=prefix)
        if model.is_enriched:
            enriched_models[identifier] = model.to_json()
        else:
            unenriched_models[identifier] = model

    logger.info(
        "%d of %d %s are identifiers already enriched",
        len(enriched_models),
        len(identifiers),
        prefix,
    )
    manager.session.commit()

    errors = set()
    if not unenriched_models or offline:
        return enriched_models, errors

    it = tqdm(unenriched_models, desc=f"getting {prefix} data in chunks of {group_size}")
    for identifier_chunk in chunked(it, n=group_size):
        helper(
            identifier_chunk,
            manager=manager,
            enriched_models=enriched_models,
            unenriched_models=unenriched_models,
            errors=errors,
        )

    return enriched_models, errors


def _help_enrich_pmids(identifiers: Iterable[str], *, manager, unenriched_models, enriched_models, errors):
    response = get_pubmed_citation_response(identifiers)
    response_pmids = response["result"]["uids"]

    for pmid in response_pmids:
        p = response["result"][pmid]
        citation = unenriched_models.get(pmid)
        if citation is None:
            tqdm.write(f"problem looking up pubmed:{pmid}")
            continue

        successful_enrichment = enrich_citation_model(manager, citation, p)

        if not successful_enrichment:
            tqdm.write(f"Error downloading pubmed:{pmid}")
            errors.add(pmid)
            continue

        enriched_models[pmid] = citation.to_json()
        manager.session.add(citation)

    manager.session.commit()  # commit in groups


def _help_enrich_pmc_identifiers(
    identifiers: Iterable[str],
    *,
    manager: Manager,
    unenriched_models,
    enriched_models,
    errors,
):
    for pmcid in identifiers:
        try:
            csl = get_pmc_csl_item(pmcid)
        except Exception:
            tqdm.write(f"Error downloading pmc:{pmcid}")
            errors.add(pmcid)
            continue

        model = unenriched_models[pmcid]
        enrich_citation_model_from_pmc(manager=manager, citation=model, csl=csl)
        manager.session.add(model)
        enriched_models[pmcid] = model.to_json()

    manager.session.commit()  # commit in groups


_HELPERS = {
    "pubmed": _help_enrich_pmids,
    "pmc": _help_enrich_pmc_identifiers,
}


def _get_citation_models(
    identifiers: Iterable[str],
    *,
    prefix: str,
    manager: Manager,
    chunksize: int = 200,
) -> Iterable[models.Citation]:
    for identifiers_chunk in chunked(identifiers, chunksize):
        citation_filter = and_(
            models.Citation.db == prefix,
            models.Citation.db_id.in_(identifiers_chunk),
        )
        yield from manager.session.query(models.Citation).filter(citation_filter).all()


def enrich_pubmed_citations(
    graph: BELGraph,
    *,
    manager: Optional[Manager] = None,
    group_size: Optional[int] = None,
    offline: bool = False,
) -> Set[str]:
    """Overwrite all PubMed citations with values from NCBI's eUtils lookup service.

    :param graph: A BEL graph
    :param manager: A PyBEL database manager
    :param group_size: The number of PubMed identifiers to query at a time. Defaults to 200 identifiers.
    :param offline: An override for when you don't want to hit the eUtils
    :return: A set of PMIDs for which the eUtils service crashed
    """
    return _enrich_citations(
        manager=manager,
        graph=graph,
        group_size=group_size,
        offline=offline,
        prefix="pubmed",
    )


def enrich_pmc_citations(
    graph: BELGraph,
    *,
    manager: Optional[Manager] = None,
    group_size: Optional[int] = None,
    offline: bool = False,
) -> Set[str]:
    """Overwrite all PubMed citations with values from NCBI's eUtils lookup service.

    :param graph: A BEL graph
    :param manager: A PyBEL database manager
    :param group_size: The number of PubMed identifiers to query at a time. Defaults to 200 identifiers.
    :param offline: An override for when you don't want to hit the eUtils
    :return: A set of PMIDs for which the eUtils service crashed
    """
    return _enrich_citations(
        manager=manager,
        graph=graph,
        group_size=group_size,
        offline=offline,
        prefix="pmc",
    )


def _enrich_citations(
    graph: BELGraph,
    manager: Optional[Manager],
    group_size: Optional[int] = None,
    offline: bool = False,
    prefix: Optional[str] = None,
) -> Set[str]:
    """Overwrite all citations of the given prefix using the predefined lookup functions.

    :param graph: A BEL Graph
    :param group_size: The number of identifiers to query at a time. Defaults to 200 identifiers.
    :return: A set of identifiers for which lookup was not possible
    """
    if manager is None:
        manager = Manager()
    if prefix is None:
        prefix = "pubmed"

    identifiers = {identifier for identifier in get_citation_identifiers(graph, prefix) if identifier}
    identifier_map, errors = _get_citations_by_identifiers(
        manager,
        identifiers=identifiers,
        group_size=group_size,
        offline=offline,
        prefix=prefix,
    )

    for u, v, k in filter_edges(graph, CITATION_PREDICATES[prefix]):
        identifier = graph[u][v][k][CITATION].identifier

        identifier_data = identifier_map.get(identifier)
        if identifier_data is None:
            logger.warning("Missing data for %s:%s", prefix, identifier)
            errors.add(identifier)
            continue

        graph[u][v][k][CITATION].update(identifier_data)

    return errors


@lru_cache()
def get_pmc_csl_item(pmcid: str) -> Mapping[str, Any]:
    """Get the CSL Item for a PubMed Central record by its PMID, PMCID, or DOI, using the NCBI Citation Exporter API."""
    if not pmcid.startswith("PMC"):
        raise ValueError(f"not a valid pmd id: {pmcid}")

    from manubot.cite.pubmed import get_pmc_csl_item

    csl_item = get_pmc_csl_item(pmcid)
    if "URL" not in csl_item:
        csl_item["URL"] = f"https://www.ncbi.nlm.nih.gov/pmc/articles/{csl_item.get('PMCID', pmcid)}/"
    return csl_item


def enrich_citation_model_from_pmc(manager: Manager, citation: models.Citation, csl: Mapping[str, Any]) -> bool:
    """Enrich a citation model with the information from PubMed Central.

    :param manager: A database manager
    :param citation: A citation model
    :param dict csl: The dictionary from PMC
    """
    citation.title = csl.get("title")
    citation.journal = csl.get("container-title")
    citation.volume = csl.get("volume")
    # citation.issue = csl['issue']
    citation.pages = csl.get("page")
    citation.article_type = csl.get("type")

    for author in csl.get("author", []):
        try:
            author_name = f'{author["given"]} {author["family"]}'
        except KeyError:
            print(f"problem with author in pmc:{citation.db_id}", author)
            continue
        author_model = manager.get_or_create_author(author_name)
        if author_model not in citation.authors:
            citation.authors.append(author_model)

    if citation.authors:
        citation.first = citation.authors[0]
        citation.last = citation.authors[-1]

    issued = csl.get("issued")
    if issued is not None:
        date_parts = issued["date-parts"][0]
        if len(date_parts) == 3:
            citation.date = date(year=date_parts[0], month=date_parts[1], day=date_parts[2])
        elif len(date_parts) == 2:
            citation.date = date(year=date_parts[0], month=date_parts[1], day=1)
        elif len(date_parts) == 1:
            citation.date = date(year=date_parts[0], month=1, day=1)
        else:
            logger.warning("not sure about date parts: %s", date_parts)

    return True



# -*- coding: utf-8 -*-

"""Exceptions for the manager."""

from ..constants import LINE
from ..exceptions import PyBELWarning

MSG = (
    "Error adding edge {line_s} to database. Check this line in the file and make sure the citation, "
    "evidence, and annotations all use valid UTF-8 characters: {source} {target} {key} {data} with "
    "original error:\n {error}"
)


class EdgeAddError(PyBELWarning):
    """When there's a problem inserting an edge."""

    def __init__(self, e, u, v, key, data):  # noqa: D107
        super().__init__(e, u, v, key, data)
        self.error = e
        self.source = u
        self.target = v
        self.key = key
        self.data = data

    def __str__(self):
        line_s = "from line {} ".format(self.line) if LINE in self.data else ""
        return MSG.format(
            line_s=line_s,
            source=self.source,
            target=self.target,
            key=self.key,
            data=self.data,
            error=self.error,
        )

    @property
    def line(self) -> str:
        """Return the BEL script's line on which this error occurred."""
        return self.data.get(LINE)



# -*- coding: utf-8 -*-

"""The :mod:`pybel.parser` module contains utilities for parsing BEL documents and BEL statements."""

from .modifiers import *
from .parse_bel import BELParser
from .parse_concept import ConceptParser
from .parse_control import ControlParser
from .parse_metadata import MetadataParser



# -*- coding: utf-8 -*-

"""This module assists in running complex workflows on BEL graphs."""

from . import decorators, exc, pipeline
from .decorators import *
from .exc import *
from .pipeline import *

__all__ = [k for k in locals() if not k.startswith("_")]



# -*- coding: utf-8 -*-

from pybel.examples import sialic_acid_graph
from pybel.examples.sialic_acid_example import (
    cd33,
    cd33_phosphorylated,
    shp2,
    syk,
    trem2,
)
from pybel.manager.models import Edge, Namespace, Network
from pybel.manager.query_manager import graph_from_edges
from pybel.resources import CHEBI_URL, GO_URL, HGNC_URL
from pybel.testing.cases import TemporaryCacheClsMixin
from pybel.testing.mocks import mock_bel_resources


class TestSeeding(TemporaryCacheClsMixin):
    """This module tests the seeding functions in the query manager."""

    @classmethod
    def setUpClass(cls):
        """Add the Sialic Acid subgraph for all query tests."""
        super().setUpClass()

        with mock_bel_resources:
            cls.manager.insert_graph(sialic_acid_graph)

    def test_namespace_existence(self):
        """Check the sialic acid graph has the right namespaces, and they're uploaded properly."""
        ns = self.manager.session.query(Namespace).filter(Namespace.url == HGNC_URL).one()
        self.assertIsNotNone(ns)

        ns = self.manager.session.query(Namespace).filter(Namespace.url == CHEBI_URL).one()
        self.assertIsNotNone(ns)

        ns = self.manager.session.query(Namespace).filter(Namespace.url == GO_URL).one()
        self.assertIsNotNone(ns)

    def test_sialic_acid_in_node_store(self):
        r = "sialic acid"

        n = self.manager.get_namespace_entry(CHEBI_URL, r)
        self.assertIsNotNone(n)

        self.assertEqual(r, n.name)

    def test_network_existence(self):
        networks = self.manager.session.query(Network).all()

        self.assertEqual(1, len(networks))

    def test_edge_existence(self):
        edges = self.manager.session.query(Edge).all()

        self.assertEqual(11, len(edges))

    def test_seed_by_pmid(self):
        pmids = ["26438529"]
        edges = self.manager.query_edges_by_pubmed_identifiers(pmids)
        self.assertLessEqual(1, len(edges))

    def test_seed_by_pmid_no_result(self):
        missing_pmids = ["11111"]
        edges = self.manager.query_edges_by_pubmed_identifiers(missing_pmids)
        self.assertEqual(0, len(edges))

    def test_seed_by_induction_raise(self):
        """Test that seeding by induction fails when an empty list is given."""
        with self.assertRaises(ValueError):
            self.manager.query_induction([])

    def test_seed_by_induction_raise_length_one(self):
        """Test that seeding by induction fails when a list of length one is given."""
        shp2_model = self.manager.get_node_by_dsl(shp2)

        with self.assertRaises(ValueError):
            self.manager.query_induction([shp2_model])

    def test_seed_by_induction(self):
        """Test seeding by inducing over a list of nodes."""
        shp2_model = self.manager.get_node_by_dsl(shp2)
        self.assertIsNotNone(shp2_model)
        syk_model = self.manager.get_node_by_dsl(syk)
        self.assertIsNotNone(syk_model)
        trem2_model = self.manager.get_node_by_dsl(trem2)
        self.assertIsNotNone(trem2_model)

        edges = self.manager.query_induction([shp2_model, syk_model, trem2_model])
        self.assertEqual(2, len(edges))

        graph = graph_from_edges(edges)

        self.assertEqual(3, graph.number_of_nodes())

        self.assertIn(trem2, graph)
        self.assertIn(syk, graph)
        self.assertIn(shp2, graph)

        self.assertEqual(3, graph.number_of_nodes())
        self.assertEqual(2, graph.number_of_edges())

    def test_seed_by_neighbors(self):
        """Test seeding a graph by neighbors of a list of nodes."""
        shp2_model = self.manager.get_node_by_dsl(shp2)
        self.assertIsNotNone(shp2_model)

        edges = self.manager.query_neighbors([shp2_model])
        self.assertEqual(2, len(edges))

        graph = graph_from_edges(edges)
        self.assertEqual(3, graph.number_of_edges())

        self.assertEqual(4, graph.number_of_nodes())

        self.assertIn(cd33_phosphorylated, graph)
        self.assertIn(cd33, graph)
        self.assertIn(syk, graph)
        self.assertIn(shp2, graph)

        self.assertEqual(3, graph.number_of_edges())



# -*- coding: utf-8 -*-

"""Importer for Hetionet JSON."""

import bz2
import json
import logging
from typing import Any, Mapping, Set, Tuple, Union

import pystow
from tqdm.autonotebook import tqdm

from .constants import (
    ACTIVATES_ACTIONS,
    BINDS_ACTIONS,
    COMPOUND,
    DSL_MAP,
    GENE,
    HETIONET_PUBMED,
    INHIBITS_ACTIONS,
    PHARMACOLOGICAL_CLASS,
    QUALIFIED_MAPPING,
    REGULATES_ACTIONS,
    UNQUALIFIED_MAPPING,
)
from ...dsl import Abundance, Protein
from ...struct import BELGraph

__all__ = [
    "get_hetionet",
    "from_hetionet_json",
    "from_hetionet_gz",
    "from_hetionet_file",
]

logger = logging.getLogger(__name__)

JSON_BZ2_URL = "https://github.com/hetio/hetionet/raw/master/hetnet/json/hetionet-v1.0.json.bz2"


def get_hetionet() -> BELGraph:
    """Get Hetionet from GitHub, cache, and convert to BEL."""
    path = pystow.ensure("bio2bel", "hetionet", url=JSON_BZ2_URL)
    return from_hetionet_gz(path.as_posix())


def from_hetionet_gz(path: str) -> BELGraph:
    """Get Hetionet from its JSON GZ file."""
    logger.info("opening %s", path)
    with bz2.open(path) as file:
        return from_hetionet_file(file)


def from_hetionet_file(file) -> BELGraph:
    """Get Hetionet from a JSON file."""
    logger.info("parsing json from %s", file)
    j = json.load(file)
    logger.info("converting hetionet dict to BEL")
    return from_hetionet_json(j)


def from_hetionet_json(
    hetionet_dict: Mapping[str, Any],
    use_tqdm: bool = True,
) -> BELGraph:
    """Convert a Hetionet dictionary to a BEL graph."""
    graph = BELGraph(  # FIXME what metadata is appropriate?
        name="Hetionet",
        version="1.0",
        authors="Daniel Himmelstein",
    )
    # FIXME add namespaces
    # graph.namespace_pattern.update({})

    kind_identifier_to_name = {(x["kind"], x["identifier"]): x["name"] for x in hetionet_dict["nodes"]}

    edges = hetionet_dict["edges"]

    if use_tqdm:
        edges = tqdm(edges, desc="Converting Hetionet", unit_scale=True)
        it_logger = edges.write
    else:
        it_logger = logger.info

    for edge in edges:
        _add_edge(graph, edge, kind_identifier_to_name, it_logger)

    return graph


def _get_node(edge, key, kind_identifier_to_name) -> Union[Tuple[None, None, None, None], Tuple[str, str, str, str]]:
    node_type, node_identifier = edge[key]
    namespace = DSL_MAP.get(node_type)
    if namespace is None:
        return None, None, None, None
    node_name = kind_identifier_to_name[node_type, node_identifier]
    node_identifier = str(node_identifier)

    if node_identifier.lower().startswith(namespace):
        node_identifier = node_identifier[1 + len(namespace) :]  # remove redundant prefix

    return node_type, namespace, node_identifier, node_name


def _add_edge(  # noqa: C901
    graph,
    edge,
    kind_identifier_to_name,
    it_logger,
) -> Union[None, str, Set[str]]:
    source_type, source_ns, source_identifier, source_name = _get_node(edge, "source_id", kind_identifier_to_name)
    target_type, target_ns, target_identifier, target_name = _get_node(edge, "target_id", kind_identifier_to_name)
    if source_type is None or target_type is None:
        return

    kind = edge["kind"]

    # direction = e['direction']
    data = edge["data"]
    if "unbiased" in data:
        del data["unbiased"]

    annotations = {}
    if "source" in data:
        source = data.pop("source")
        annotations["source"] = {source: True}
    elif "sources" in data:
        annotations["source"] = {source: True for source in data.pop("sources")}
    else:
        pass
        # it_logger(f'Missing source for {source_identifier}-{kind}-{target_identifier}\n{e}')

    if "pubmed_ids" in data:
        citations = list(data.pop("pubmed_ids"))
    else:
        citations = [HETIONET_PUBMED]

    for k, v in data.items():
        if k in {"actions", "urls", "subtypes"}:
            continue  # handled explicitly later
        if not isinstance(v, (str, int, bool, float)):
            it_logger(
                "Unhandled: {source_identifier}-{kind}-{target_identifier} {k}: {v}".format(
                    source_identifier=source_identifier,
                    kind=kind,
                    target_identifier=target_identifier,
                    k=k,
                    v=v,
                )
            )
            continue
        annotations[k] = {v: True}

    for _h_type, h_dsl, _r, _t_type, t_dsl, f in QUALIFIED_MAPPING:
        if source_type != _h_type or kind != _r or target_type != _t_type:
            continue
        rv = set()
        for citation in citations:
            key = f(
                graph,
                h_dsl(namespace=source_ns, identifier=source_identifier, name=source_name),
                t_dsl(namespace=target_ns, identifier=target_identifier, name=target_name),
                citation=citation,
                evidence="",
                annotations=annotations,
            )
            rv.add(key)
        return rv

    for _h_type, h_dsl, _r, _t_type, t_dsl, f in UNQUALIFIED_MAPPING:
        if source_type == _h_type and kind == _r and target_type == _t_type:
            return f(
                graph,
                h_dsl(namespace=source_ns, identifier=source_identifier, name=source_name),
                t_dsl(namespace=target_ns, identifier=target_identifier, name=target_name),
            )

    def _check(_source_type: str, _kind: str, _target_type: str) -> bool:
        """Check the metaedge."""
        return kind == _kind and source_type == _source_type and target_type == _target_type

    if _check(COMPOUND, "binds", GENE):
        drug = Abundance(namespace="drugbank", name=source_name, identifier=source_identifier)
        protein = Protein(namespace="ncbigene", name=target_name, identifier=target_identifier)

        rv = set()
        for action in data.get("actions", []):
            action = action.lower()
            if action in ACTIVATES_ACTIONS:
                key = graph.add_directly_activates(
                    drug,
                    protein,
                    citation=HETIONET_PUBMED,
                    evidence="",
                    annotations=annotations,
                )
            elif action in INHIBITS_ACTIONS:
                key = graph.add_directly_inhibits(
                    drug,
                    protein,
                    citation=HETIONET_PUBMED,
                    evidence="",
                    annotations=annotations,
                )
            elif action in REGULATES_ACTIONS:
                key = graph.add_regulates(
                    drug,
                    protein,
                    citation=HETIONET_PUBMED,
                    evidence="",
                    annotations=annotations,
                )
            elif action in BINDS_ACTIONS:
                key = graph.add_binds(
                    drug,
                    protein,
                    citation=HETIONET_PUBMED,
                    evidence="",
                    annotations=annotations,
                )
            else:
                key = graph.add_binds(
                    drug,
                    protein,
                    citation=HETIONET_PUBMED,
                    evidence="",
                    annotations=annotations,
                )
                it_logger(
                    "Unhandled action for {source_identifier}-{kind}-{target_identifier}: {action}".format(
                        source_identifier=source_identifier,
                        kind=kind,
                        target_identifier=target_identifier,
                        action=action,
                    )
                )
            rv.add(key)
        return rv

    if _check(PHARMACOLOGICAL_CLASS, "includes", COMPOUND):
        return graph.add_is_a(
            Abundance(namespace="drugbank", name=target_name, identifier=target_identifier),
            Abundance(namespace="drugcentral", name=source_name, identifier=source_identifier),
        )

    it_logger("missed: {edge}".format(edge=edge))



# -*- coding: utf-8 -*-

import logging
import unittest
from io import StringIO

from pybel import BELGraph
from pybel.examples.egf_example import egf_graph
from pybel.struct.mutation import enrich_protein_and_rna_origins
from pybel.struct.pipeline import Pipeline
from pybel.struct.pipeline.decorators import get_transformation, mapped
from pybel.struct.pipeline.exc import MetaValueError, MissingPipelineFunctionError

log = logging.getLogger(__name__)
log.setLevel(10)


class TestEgfExample(unittest.TestCase):
    """Random test for mutation functions"""

    def setUp(self):
        self.graph = egf_graph.copy()
        self.original_number_nodes = self.graph.number_of_nodes()
        self.original_number_edges = self.graph.number_of_edges()

    def check_original_unchanged(self):
        self.assertEqual(
            self.original_number_nodes,
            self.graph.number_of_nodes(),
            msg="original graph nodes should remain unchanged",
        )
        self.assertEqual(
            self.original_number_edges,
            self.graph.number_of_edges(),
            msg="original graph edges should remain unchanged",
        )


class TestPipelineFailures(unittest.TestCase):
    def test_assert_failure(self):
        with self.assertRaises(MissingPipelineFunctionError):
            get_transformation("missing function")

    def test_assert_success(self):
        m = list(mapped)
        self.assertLess(0, len(m))
        m = m[0]
        f = get_transformation(m)
        self.assertIsNotNone(f)

    def test_append_invalid(self):
        """Test when an invalid type is given to a :class:`pybel.struct.Pipeline`."""
        p = Pipeline()
        with self.assertRaises(TypeError):
            p.append(4)

    def test_get_function_failure(self):
        p = Pipeline()

        with self.assertRaises(MissingPipelineFunctionError):
            p._get_function("nonsense name")

    def test_build_meta_failure(self):
        p1, p2 = Pipeline(), Pipeline()

        p = Pipeline._build_meta("wrong", [p1, p2])

        with self.assertRaises(MetaValueError):
            p(BELGraph())

    def test_fail_add(self):
        pipeline = Pipeline()
        with self.assertRaises(MissingPipelineFunctionError):
            pipeline.append("missing function")


class TestPipeline(TestEgfExample):
    def test_append(self):
        pipeline = Pipeline()
        self.assertEqual(0, len(pipeline))

        pipeline.append("enrich_protein_and_rna_origins")
        self.assertEqual(1, len(pipeline))

    def test_extend(self):
        p1 = Pipeline.from_functions(["enrich_protein_and_rna_origins"])
        self.assertEqual(1, len(p1))

        p2 = Pipeline.from_functions(["remove_pathologies"])
        p1.extend(p2)

        self.assertEqual(2, len(p1))

    def test_serialize_string(self):
        p = Pipeline.from_functions(["enrich_protein_and_rna_origins"])
        s = p.dumps()
        p_reconstituted = Pipeline.loads(s)
        self.assertEqual(p.protocol, p_reconstituted.protocol)

    def test_serialize_file(self):
        p = Pipeline.from_functions(["enrich_protein_and_rna_origins"])
        sio = StringIO()
        p.dump(sio)
        sio.seek(0)
        p_reconstituted = Pipeline.load(sio)
        self.assertEqual(p.protocol, p_reconstituted.protocol)

    def test_pipeline_by_string(self):
        pipeline = Pipeline.from_functions(
            [
                "enrich_protein_and_rna_origins",
            ]
        )
        result = pipeline(self.graph)

        self.assertEqual(32, result.number_of_nodes())

        for node in self.graph:
            self.assertIn(node, result)

        self.check_original_unchanged()

    def test_pipeline_by_function(self):
        pipeline = Pipeline.from_functions(
            [
                enrich_protein_and_rna_origins,
            ]
        )
        result = pipeline(self.graph)

        self.assertEqual(32, result.number_of_nodes())

        for node in self.graph:
            self.assertIn(node, result)

        self.check_original_unchanged()



# -*- coding: utf-8 -*-

"""Contains the main data structure for PyBEL."""

import logging
import warnings
from collections import Counter, defaultdict
from copy import deepcopy
from functools import partialmethod
from itertools import chain
from textwrap import dedent
from typing import (
    Any,
    Dict,
    Hashable,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    TextIO,
    Tuple,
    Union,
)

import networkx as nx
from tabulate import tabulate

from .operations import left_full_join, left_node_intersection_join, left_outer_join
from .utils import update_metadata
from ..canonicalize import edge_to_bel
from ..constants import (
    ACTIVITY,
    ANNOTATIONS,
    ASSOCIATION,
    CAUSES_NO_CHANGE,
    CITATION,
    CITATION_AUTHORS,
    CITATION_TYPE_PUBMED,
    CORRELATION,
    DECREASES,
    DEGRADATION,
    DIRECTLY_DECREASES,
    DIRECTLY_INCREASES,
    DIRECTLY_REGULATES,
    EFFECT,
    EQUIVALENT_TO,
    EVIDENCE,
    FROM_LOC,
    GRAPH_ANNOTATION_CURIE,
    GRAPH_ANNOTATION_LIST,
    GRAPH_ANNOTATION_MIRIAM,
    GRAPH_ANNOTATION_PATTERN,
    GRAPH_ANNOTATION_URL,
    GRAPH_METADATA,
    GRAPH_NAMESPACE_PATTERN,
    GRAPH_NAMESPACE_URL,
    GRAPH_PATH,
    GRAPH_PYBEL_VERSION,
    HAS_PRODUCT,
    HAS_REACTANT,
    HAS_VARIANT,
    IDENTIFIER,
    INCREASES,
    IS_A,
    LOCATION,
    METADATA_AUTHORS,
    METADATA_CONTACT,
    METADATA_COPYRIGHT,
    METADATA_DESCRIPTION,
    METADATA_DISCLAIMER,
    METADATA_LICENSES,
    METADATA_NAME,
    METADATA_VERSION,
    MODIFIER,
    NAMESPACE,
    NEGATIVE_CORRELATION,
    NO_CORRELATION,
    ORTHOLOGOUS,
    PART_OF,
    POSITIVE_CORRELATION,
    REGULATES,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    TO_LOC,
    TRANSCRIBED_TO,
    TRANSLATED_TO,
    TRANSLOCATION,
)
from ..dsl import (
    BaseAbundance,
    BaseConcept,
    BaseEntity,
    CentralDogma,
    ComplexAbundance,
    Gene,
    ListAbundance,
    MicroRna,
    Protein,
    ProteinModification,
    Reaction,
    Rna,
    activity,
)
from ..exceptions import BELParserWarning
from ..language import CitationDict, Entity, citation_dict
from ..typing import EdgeData
from ..utils import hash_edge
from ..version import get_version

__all__ = [
    "BELGraph",
]

logger = logging.getLogger(__name__)

AnnotationsDict = Mapping[str, List[Entity]]
AnnotationsHint = Union[Mapping[str, str], Mapping[str, Set[str]], AnnotationsDict]
WarningTuple = Tuple[Optional[str], BELParserWarning, EdgeData]


class BELGraph(nx.MultiDiGraph):
    """An extension to :class:`networkx.MultiDiGraph` to represent BEL."""

    #: A set of pairs of hashes of edges over which there is transitivity.
    #: For example, for the nested statement (P(X) -> P(Y)) -> P(Z) will have
    #: a pair for (hash(P(X) -> P(Y)), hash(P(Y) -> P(Z)))
    transitivities: Set[Tuple[str, str]]

    def __init__(
        self,
        name: Optional[str] = None,
        version: Optional[str] = None,
        description: Optional[str] = None,
        authors: Optional[str] = None,
        contact: Optional[str] = None,
        license: Optional[str] = None,
        copyright: Optional[str] = None,
        disclaimer: Optional[str] = None,
        path: Optional[str] = None,
    ) -> None:
        """Initialize a BEL graph with its associated metadata.

        :param name: The graph's name
        :param version: The graph's version. Recommended to use `semantic versioning <http://semver.org/>`_ or
         ``YYYYMMDD`` format.
        :param description: A description of the graph
        :param authors: The authors of this graph
        :param contact: The contact email for this graph
        :param license: The license for this graph
        :param copyright: The copyright for this graph
        :param disclaimer: The disclaimer for this graph
        """
        super().__init__()

        self._warnings = []

        self.graph.update(
            {
                GRAPH_PYBEL_VERSION: get_version(),
                GRAPH_METADATA: {},
                GRAPH_NAMESPACE_URL: {},
                GRAPH_NAMESPACE_PATTERN: {},
                GRAPH_ANNOTATION_URL: {},
                GRAPH_ANNOTATION_PATTERN: {},
                GRAPH_ANNOTATION_LIST: defaultdict(set),
                GRAPH_ANNOTATION_CURIE: set(),
                GRAPH_ANNOTATION_MIRIAM: set(),
            }
        )

        if name:
            self.name = name

        if version:
            self.version = version

        if description:
            self.description = description

        if authors:
            self.authors = authors

        if contact:
            self.contact = contact

        if license:
            self.license = license

        if copyright:
            self.copyright = copyright

        if disclaimer:
            self.disclaimer = disclaimer

        if path:
            self.path = path

        self.transitivities = set()

        #: A reference to the parent graph
        self.parent = None
        self._count = CountDispatch(self)
        self._expand = ExpandDispatch(self)
        self._induce = InduceDispatch(self)
        self._plot = PlotDispatch(self)
        self._summary = SummarizeDispatch(self)

        self.raise_on_missing_annotations = True

    def child(self) -> "BELGraph":
        """Create an empty graph with a "parent" reference back to this one."""
        rv = BELGraph()
        rv.parent = self
        update_metadata(source=self, target=rv)
        return rv

    @property
    def count(self) -> "CountDispatch":  # noqa: D401
        """A dispatch to count functions.

        Can be used like this:

        >>> from pybel.examples import sialic_acid_graph
        >>> sialic_acid_graph.count.functions()
        Counter({'Protein': 7, 'Complex': 1, 'Abundance': 1})
        """
        return self._count

    @property
    def summarize(self) -> "SummarizeDispatch":  # noqa: D401
        """A dispatch to summarize the graph."""
        return self._summary

    def _repr_html_(self):
        return self._summary._repr_html_()

    @property
    def expand(self) -> "ExpandDispatch":  # noqa: D401
        """A dispatch to expand the graph w.r.t. its parent."""
        return self._expand

    @property
    def induce(self) -> "InduceDispatch":  # noqa: D401
        """A dispatch to mutate the graph."""
        return self._induce

    @property
    def plot(self) -> "PlotDispatch":  # noqa: D401
        """A dispatch to plot the graph using :mod:`matplotlib` and :mod:`seaborn`."""
        return self._plot

    @property
    def path(self) -> Optional[str]:  # noqa: D401
        """The graph's path, if it was derived from a BEL document."""
        return self.graph.get(GRAPH_PATH)

    @path.setter
    def path(self, path: str) -> None:
        """Set the graph's path."""
        self.graph[GRAPH_PATH] = path

    @property
    def document(self) -> Dict[str, Any]:  # noqa: D401
        """The dictionary holding the metadata from the ``SET DOCUMENT`` statements in the source BEL script.

        All keys are normalized according to :data:`pybel.constants.DOCUMENT_KEYS`.
        """
        return self.graph[GRAPH_METADATA]

    @property
    def name(self, *attrs) -> Optional[str]:  # noqa: D401 # Needs *attrs since it's an override
        """The graph's name.

        .. hint:: Can be set with the ``SET DOCUMENT Name = "..."`` entry in the source BEL script.
        """
        return self.document.get(METADATA_NAME)

    @name.setter
    def name(self, *attrs, **kwargs):  # Needs *attrs and **kwargs since it's an override
        """Set the graph's name."""
        self.document[METADATA_NAME] = attrs[0]

    @property
    def version(self) -> Optional[str]:  # noqa: D401
        """The graph's version.

        .. hint:: Can be set with the ``SET DOCUMENT Version = "..."`` entry in the source BEL script.
        """
        return self.document.get(METADATA_VERSION)

    @version.setter
    def version(self, version):
        """Set the graph's version."""
        self.document[METADATA_VERSION] = version

    @property
    def description(self) -> Optional[str]:  # noqa: D401
        """The graph's description.

        .. hint:: Can be set with the ``SET DOCUMENT Description = "..."`` entry in the source BEL document.
        """
        return self.document.get(METADATA_DESCRIPTION)

    @description.setter
    def description(self, description: str) -> None:
        """Set the graph's description."""
        self.document[METADATA_DESCRIPTION] = description

    @property
    def authors(self) -> Optional[str]:  # noqa: D401
        """The graph's authors.

        .. hint:: Can be set with the ``SET DOCUMENT Authors = "..."`` entry in the source BEL document.
        """
        return self.document.get(METADATA_AUTHORS)

    @authors.setter
    def authors(self, authors: str) -> None:
        """Set the graph's authors."""
        self.document[METADATA_AUTHORS] = authors

    @property
    def contact(self) -> Optional[str]:  # noqa: D401
        """The graph's contact information.

        .. hint:: Can be set with the ``SET DOCUMENT ContactInfo = "..."`` entry in the source BEL document.
        """
        return self.document.get(METADATA_CONTACT)

    @contact.setter
    def contact(self, contact: str) -> None:
        """Set the graph's contact."""
        self.document[METADATA_CONTACT] = contact

    @property
    def license(self) -> Optional[str]:  # noqa: D401
        """The graph's license.

        .. hint:: Can be set with the ``SET DOCUMENT Licenses = "..."`` entry in the source BEL document
        """
        return self.document.get(METADATA_LICENSES)

    @license.setter
    def license(self, license_str: str) -> None:
        """Set the graph's license."""
        self.document[METADATA_LICENSES] = license_str

    @property
    def copyright(self) -> Optional[str]:  # noqa: D401
        """The graph's copyright.

        .. hint:: Can be set with the ``SET DOCUMENT Copyright = "..."`` entry in the source BEL document
        """
        return self.document.get(METADATA_COPYRIGHT)

    @copyright.setter
    def copyright(self, copyright_str: str) -> None:
        """Set the graph's copyright."""
        self.document[METADATA_COPYRIGHT] = copyright_str

    @property
    def disclaimer(self) -> Optional[str]:  # noqa: D401
        """The graph's disclaimer.

        .. hint:: Can be set with the ``SET DOCUMENT Disclaimer = "..."`` entry in the source BEL document.
        """
        return self.document.get(METADATA_DISCLAIMER)

    @disclaimer.setter
    def disclaimer(self, disclaimer: str) -> None:
        """Set the graph's disclaimer."""
        self.document[METADATA_DISCLAIMER] = disclaimer

    @property
    def namespace_url(self) -> Dict[str, str]:  # noqa: D401
        """The mapping from the keywords used in this graph to their respective BEL namespace URLs.

        .. hint:: Can be appended with the ``DEFINE NAMESPACE [key] AS URL "[value]"`` entries in the definitions
                  section of the source BEL document.
        """
        return self.graph[GRAPH_NAMESPACE_URL]

    @property
    def defined_namespace_keywords(self) -> Set[str]:  # noqa: D401
        """The set of all keywords defined as namespaces in this graph."""
        return set(self.namespace_pattern) | set(self.namespace_url)

    @property
    def namespace_pattern(self) -> Dict[str, str]:  # noqa: D401
        """The mapping from the namespace keywords used to create this graph to their regex patterns.

        .. hint:: Can be appended with the ``DEFINE NAMESPACE [key] AS PATTERN "[value]"`` entries in the definitions
                  section of the source BEL document.
        """
        return self.graph[GRAPH_NAMESPACE_PATTERN]

    @property
    def annotation_url(self) -> Dict[str, str]:  # noqa: D401
        """The mapping from the annotation keywords used to create this graph to the URLs of the BELANNO files.

        .. hint:: Can be appended with the ``DEFINE ANNOTATION [key] AS URL "[value]"`` entries in the definitions
                  section of the source BEL document.
        """
        return self.graph[GRAPH_ANNOTATION_URL]

    @property
    def annotation_miriam(self) -> Set[str]:  # noqa: D401
        """The set of annotations defined by MIRIAM."""
        if GRAPH_ANNOTATION_MIRIAM not in self.graph:
            self.graph[GRAPH_ANNOTATION_MIRIAM] = set()
        return self.graph[GRAPH_ANNOTATION_MIRIAM]

    @property
    def annotation_curie(self) -> Set[str]:  # noqa: D401
        """The set of annotations defined by CURIE."""
        if GRAPH_ANNOTATION_CURIE not in self.graph:
            self.graph[GRAPH_ANNOTATION_CURIE] = set()
        return self.graph[GRAPH_ANNOTATION_CURIE]

    @property
    def annotation_pattern(self) -> Dict[str, str]:  # noqa: D401
        """The mapping from the annotation keywords used to create this graph to their regex patterns as strings.

        .. hint:: Can be appended with the ``DEFINE ANNOTATION [key] AS PATTERN "[value]"`` entries in the definitions
                  section of the source BEL document.
        """
        return self.graph[GRAPH_ANNOTATION_PATTERN]

    @property
    def annotation_list(self) -> Dict[str, Set[str]]:  # noqa: D401
        """The mapping from the keywords of locally defined annotations to their respective sets of values.

        .. hint:: Can be appended with the ``DEFINE ANNOTATION [key] AS LIST {"[value]", ...}`` entries in the
                  definitions section of the source BEL document.
        """
        return self.graph[GRAPH_ANNOTATION_LIST]

    @property
    def defined_annotation_keywords(self) -> Set[str]:
        """Get the set of all keywords defined as annotations in this graph."""
        return set(self.annotation_pattern) | set(self.annotation_url) | set(self.annotation_list)

    @property
    def pybel_version(self) -> str:  # noqa: D401
        """The version of PyBEL with which this graph was produced as a string."""
        return self.graph[GRAPH_PYBEL_VERSION]

    @property
    def warnings(self) -> List[WarningTuple]:  # noqa: D401
        """A list of warnings associated with this graph."""
        return self._warnings

    def number_of_warnings(self) -> int:
        """Return the number of warnings."""
        return len(self.warnings)

    def number_of_citations(self) -> int:
        """Return the number of citations contained within the graph."""
        return self.count.citations()

    def number_of_authors(self) -> int:
        """Return the number of authors contained within the graph."""
        return len(self.get_authors())

    def get_authors(self) -> Set[str]:
        """Get the authors for the citations in the graph."""
        return set(self.count.authors())

    def __str__(self):
        return "{} v{}".format(self.name, self.version)

    def add_transitivity(self, k1: str, k2: str) -> None:
        """Add a pair of edge hashes over which there is transitivity.

        :param k1: The hash of the subject edge
        :param k2: The hash of the object edge
        """
        self.transitivities.add((k1, k2))

    def add_warning(
        self,
        exception: BELParserWarning,
        context: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Add a warning to the internal warning log in the graph, with optional context information.

        :param exception: The exception that occurred
        :param context: The context from the parser when the exception occurred
        """
        self.warnings.append(
            (
                self.path,
                exception,
                {} if context is None else context,
            )
        )

    def _help_add_edge(self, source: BaseEntity, target: BaseEntity, attr: Mapping) -> str:
        """Help add a pre-built edge."""
        self.add_node_from_data(source)
        self.add_node_from_data(target)
        return self._help_add_edge_helper(source=source, target=target, attr=attr)

    def _help_add_edge_helper(self, source: BaseEntity, target: BaseEntity, attr: Mapping[str, Any]) -> str:
        key = hash_edge(source, target, attr)

        if not self.has_edge(source, target, key):
            self.add_edge(source, target, key=key, **attr)

        return key

    def add_unqualified_edge(self, source: BaseEntity, target: BaseEntity, relation: str) -> str:
        """Add a unique edge that has no annotations.

        :param source: The source node
        :param target: The target node
        :param relation: A relationship label from :mod:`pybel.constants`
        :return: The key for this edge (a unique hash)
        """
        attr = {RELATION: relation}
        return self._help_add_edge(source=source, target=target, attr=attr)

    def add_transcription(self, gene: Gene, rna: Union[Rna, MicroRna]) -> str:
        """Add a transcription relation from a gene to an RNA or miRNA node.

        :param gene: A gene node
        :param rna: An RNA or microRNA node
        """
        return self.add_unqualified_edge(gene, rna, TRANSCRIBED_TO)

    def add_translation(self, rna: Rna, protein: Protein) -> str:
        """Add a translation relation from a RNA to a protein.

        :param rna: An RNA node
        :param protein: A protein node
        """
        return self.add_unqualified_edge(rna, protein, TRANSLATED_TO)

    def _add_two_way_qualified_edge(self, source: BaseEntity, target: BaseEntity, *args, **kwargs) -> str:
        """Add an qualified edge both ways."""
        self.add_qualified_edge(source=target, target=source, *args, **kwargs)
        return self.add_qualified_edge(source=source, target=target, *args, **kwargs)

    def _add_two_way_unqualified_edge(self, source: BaseEntity, target: BaseEntity, *args, **kwargs) -> str:
        """Add an unqualified edge both ways."""
        self.add_unqualified_edge(target, source, *args, **kwargs)
        return self.add_unqualified_edge(source, target, *args, **kwargs)

    add_equivalence = partialmethod(_add_two_way_unqualified_edge, relation=EQUIVALENT_TO)
    """Add two equivalence relations for the nodes."""

    add_orthology = partialmethod(_add_two_way_unqualified_edge, relation=ORTHOLOGOUS)
    """Add two orthology relations for the nodes such that ``u orthologousTo v`` and ``v orthologousTo u``."""

    add_is_a = partialmethod(add_unqualified_edge, relation=IS_A)
    """Add an ``isA`` relationship such that ``u isA v``."""

    add_part_of = partialmethod(add_unqualified_edge, relation=PART_OF)
    """Add a ``partOf`` relationship such that ``u partOf v``."""

    add_has_variant = partialmethod(add_unqualified_edge, relation=HAS_VARIANT)
    """Add a ``hasVariant`` relationship such that ``u hasVariant v``."""

    add_has_reactant = partialmethod(add_unqualified_edge, relation=HAS_REACTANT)
    """Add a ``hasReactant`` relationship such that ``u hasReactant v``."""

    add_has_product = partialmethod(add_unqualified_edge, relation=HAS_PRODUCT)
    """Add a ``hasProduct`` relationship such that ``u hasProduct v``."""

    def add_qualified_edge(
        self,
        source: BaseEntity,
        target: BaseEntity,
        *,
        relation: str,
        evidence: str,
        citation: Union[str, Tuple[str, str], CitationDict],
        annotations: Optional[AnnotationsHint] = None,
        source_modifier: Optional[Mapping[str, Any]] = None,
        target_modifier: Optional[Mapping[str, Any]] = None,
        **attr,
    ) -> str:
        """Add a qualified edge.

        Qualified edges have a relation, evidence, citation, and optional annotations, subject modifications,
        and object modifications.

        :param source: The source node
        :param target: The target node
        :param relation: The type of relation this edge represents
        :param evidence: The evidence string from an article
        :param citation: The citation data dictionary for this evidence. If a string is given,
         assumes it's a PubMed identifier and auto-fills the citation type.
        :param annotations: The annotations data dictionary
        :param source_modifier: The modifiers (like activity) on the subject node. See data model documentation.
        :param target_modifier: The modifiers (like activity) on the object node. See data model documentation.

        :return: The hash of the edge
        """
        if "subject_modifier" in attr:
            warnings.warn("subject_modifier has been renamed to source_modifier")
            source_modifier = attr.pop("subject_modifier")
        if "object_modifier" in attr:
            warnings.warn("object_modifier has been renamed to target_modifier")
            target_modifier = attr.pop("object_modifier")

        attr = self._build_attr(
            relation=relation,
            evidence=evidence,
            citation=citation,
            annotations=annotations,
            source_modifier=source_modifier,
            target_modifier=target_modifier,
            **attr,
        )
        return self._help_add_edge(source=source, target=target, attr=attr)

    def _build_attr(
        self,
        relation: str,
        evidence: str,
        citation: Union[str, Tuple[str, str], CitationDict],
        annotations: Optional[AnnotationsHint] = None,
        source_modifier: Optional[Dict[str, Any]] = None,
        target_modifier: Optional[Dict[str, Any]] = None,
        **attr,
    ):
        attr.update(
            {
                RELATION: relation,
                EVIDENCE: evidence,
                CITATION: _handle_citation(citation),
            }
        )

        if annotations:  # clean up annotations
            attr[ANNOTATIONS] = self._clean_annotations(annotations)

        if source_modifier:
            attr[SOURCE_MODIFIER] = _handle_modifier(source_modifier)

        if target_modifier:
            attr[TARGET_MODIFIER] = _handle_modifier(target_modifier)

        return attr

    def add_binds(
        self,
        source: BaseAbundance,
        target: BaseAbundance,
        *,
        evidence: str,
        citation: Union[str, Tuple[str, str], CitationDict],
        annotations: Optional[AnnotationsHint] = None,
        **attr,
    ) -> str:
        """Add a "binding" relationship between the two entities such that ``u => complex(u, v)``."""
        complex_abundance = ComplexAbundance([source, target])
        return self.add_directly_increases(
            source=source,
            target=complex_abundance,
            citation=citation,
            evidence=evidence,
            annotations=annotations,
            **attr,
        )

    add_increases = partialmethod(add_qualified_edge, relation=INCREASES)
    """Wrap :meth:`add_qualified_edge` for the :data:`pybel.constants.INCREASES` relation."""

    add_directly_increases = partialmethod(add_qualified_edge, relation=DIRECTLY_INCREASES)
    """Add a :data:`pybel.constants.DIRECTLY_INCREASES` with :meth:`add_qualified_edge`."""

    add_decreases = partialmethod(add_qualified_edge, relation=DECREASES)
    """Add a :data:`pybel.constants.DECREASES` relationship with :meth:`add_qualified_edge`."""

    add_directly_decreases = partialmethod(add_qualified_edge, relation=DIRECTLY_DECREASES)
    """Add a :data:`pybel.constants.DIRECTLY_DECREASES` relationship with :meth:`add_qualified_edge`."""

    add_association = partialmethod(_add_two_way_qualified_edge, relation=ASSOCIATION)
    """Add a :data:`pybel.constants.ASSOCIATION` relationship with :meth:`add_qualified_edge`."""

    add_regulates = partialmethod(add_qualified_edge, relation=REGULATES)
    """Add a :data:`pybel.constants.REGULATES` relationship with :meth:`add_qualified_edge`."""

    add_directly_regulates = partialmethod(add_qualified_edge, relation=DIRECTLY_REGULATES)
    """Add a :data:`pybel.constants.DIRECTLY_REGULATES` relationship with :meth:`add_qualified_edge`."""

    add_correlation = partialmethod(_add_two_way_qualified_edge, relation=CORRELATION)
    """Add a :data:`pybel.constants.CORRELATION` relationship with :meth:`add_qualified_edge`."""

    add_no_correlation = partialmethod(_add_two_way_qualified_edge, relation=NO_CORRELATION)
    """Add a :data:`pybel.constants.NO_CORRELATION` relationship with :meth:`add_qualified_edge`."""

    add_positive_correlation = partialmethod(_add_two_way_qualified_edge, relation=POSITIVE_CORRELATION)
    """Add a :data:`pybel.constants.POSITIVE_CORRELATION` relationship with :meth:`add_qualified_edge`."""

    add_negative_correlation = partialmethod(_add_two_way_qualified_edge, relation=NEGATIVE_CORRELATION)
    """Add a :data:`pybel.constants.NEGATIVE_CORRELATION` relationship with :meth:`add_qualified_edge`."""

    add_causes_no_change = partialmethod(add_qualified_edge, relation=CAUSES_NO_CHANGE)
    """Add a :data:`pybel.constants.CAUSES_NO_CHANGE` relationship with :meth:`add_qualified_edge`."""

    add_inhibits = partialmethod(add_decreases, target_modifier=activity())
    """Add an "inhibits" relationship.

    A more specific version of :meth:`add_decreases` that automatically populates the object modifier with an
    activity."""

    add_directly_inhibits = partialmethod(add_directly_decreases, target_modifier=activity())

    add_activates = partialmethod(add_increases, target_modifier=activity())
    """Add an "activates" relationship.

    A more specific version of :meth:`add_increases` that automatically populates the object modifier with an
    activity."""

    add_directly_activates = partialmethod(add_directly_increases, target_modifier=activity())

    def _modify(
        self,
        add_edge_fn: str,
        name: str,
        source: BaseEntity,
        target: CentralDogma,
        code: Optional[str] = None,
        position: Optional[int] = None,
        *,
        evidence: str,
        citation: Union[str, Mapping[str, str]],
        annotations: Optional[AnnotationsHint] = None,
        source_modifier: Optional[Mapping] = None,
        target_modifier: Optional[Mapping] = None,
        **attr,
    ):
        """Add a simple modification."""
        adder = getattr(self, add_edge_fn)
        return adder(
            source=source,
            target=target.with_variants(
                ProteinModification(
                    name=name,
                    code=code,
                    position=position,
                )
            ),
            evidence=evidence,
            citation=citation,
            annotations=annotations,
            source_modifier=source_modifier,
            target_modifier=target_modifier,
            **attr,
        )

    add_phosphorylates = partialmethod(_modify, "add_increases", "Ph")
    """Add an increase of modified object with phosphorylation."""

    add_directly_phosphorylates = partialmethod(_modify, "add_directly_increases", "Ph")
    """Add a direct increase of modified object with phosphorylation."""

    add_dephosphorylates = partialmethod(_modify, "add_decreases", "Ph")
    """Add a decrease of modified object with phosphorylation."""

    add_directly_dephosphorylates = partialmethod(_modify, "add_directly_decreases", "Ph")
    """Add a direct decrease of modified object with phosphorylation."""

    def add_node_from_data(self, node: BaseEntity) -> None:
        """Add an entity to the graph."""
        assert isinstance(node, BaseEntity)

        if node in self:
            return

        self.add_node(node)

        if isinstance(node, CentralDogma) and node.variants:
            self.add_has_variant(node.get_parent(), node)

        elif isinstance(node, ListAbundance):
            for member in node.members:
                self.add_part_of(member, node)

        elif isinstance(node, Reaction):
            for reactant in node.reactants:
                self.add_has_reactant(node, reactant)
            for product in node.products:
                self.add_has_product(node, product)

    def add_reaction(
        self,
        reactants: Union[BaseAbundance, Iterable[BaseAbundance]],
        products: Union[BaseAbundance, Iterable[BaseAbundance]],
    ) -> None:
        """Add a reaction directly to the graph."""
        return self.add_node_from_data(Reaction(reactants=reactants, products=products))

    def _has_edge_attr(self, u: BaseEntity, v: BaseEntity, key: str, attr: Hashable) -> bool:
        assert isinstance(u, BaseEntity)
        assert isinstance(v, BaseEntity)
        return attr in self[u][v][key]

    def has_edge_citation(self, u: BaseEntity, v: BaseEntity, key: str) -> bool:
        """Check if the given edge has a citation."""
        return self._has_edge_attr(u, v, key, CITATION)

    def has_edge_evidence(self, u: BaseEntity, v: BaseEntity, key: str) -> bool:
        """Check if the given edge has an evidence."""
        return self._has_edge_attr(u, v, key, EVIDENCE)

    def _get_edge_attr(self, u: BaseEntity, v: BaseEntity, key: str, attr: str):
        return self[u][v][key].get(attr)

    def get_edge_citation(self, u: BaseEntity, v: BaseEntity, key: str) -> Optional[CitationDict]:
        """Get the citation for a given edge."""
        return self._get_edge_attr(u, v, key, CITATION)

    def get_edge_evidence(self, u: BaseEntity, v: BaseEntity, key: str) -> Optional[str]:
        """Get the evidence for a given edge."""
        return self._get_edge_attr(u, v, key, EVIDENCE)

    def get_edge_annotations(self, u, v, key: str) -> Optional[AnnotationsDict]:
        """Get the annotations for a given edge."""
        return self._get_edge_attr(u, v, key, ANNOTATIONS)

    def __add__(self, other: "BELGraph") -> "BELGraph":
        """Copy this graph and join it with another graph with it using :func:`pybel.struct.left_full_join`.

        :param other: Another BEL graph

        Example usage:

        >>> from pybel.examples import ras_tloc_graph, braf_graph
        >>> k = ras_tloc_graph + braf_graph
        """
        if not isinstance(other, BELGraph):
            raise TypeError("{} is not a {}".format(other, self.__class__.__name__))

        result = deepcopy(self)
        left_full_join(result, other)
        return result

    def __iadd__(self, other: "BELGraph") -> "BELGraph":
        """Join another graph into this one, in-place, using :func:`pybel.struct.left_full_join`.

        :param other: Another BEL graph

        Example usage:

        >>> from pybel.examples import ras_tloc_graph, braf_graph
        >>> ras_tloc_graph += braf_graph
        """
        if not isinstance(other, BELGraph):
            raise TypeError("{} is not a {}".format(other, self.__class__.__name__))

        left_full_join(self, other)
        return self

    def __and__(self, other: "BELGraph") -> "BELGraph":
        """Create a deep copy of this graph and left outer joins another graph.

        Uses :func:`pybel.struct.left_outer_join`.

        :param other: Another BEL graph

        Example usage:

        >>> from pybel.examples import ras_tloc_graph, braf_graph
        >>> k = ras_tloc_graph & braf_graph
        """
        if not isinstance(other, BELGraph):
            raise TypeError("{} is not a {}".format(other, self.__class__.__name__))

        result = deepcopy(self)
        left_outer_join(result, other)
        return result

    def __iand__(self, other: "BELGraph") -> "BELGraph":
        """Join another graph into this one, in-place, using :func:`pybel.struct.left_outer_join`.

        :param other: Another BEL graph

        Example usage:

        >>> from pybel.examples import ras_tloc_graph, braf_graph
        >>> ras_tloc_graph &= braf_graph
        """
        if not isinstance(other, BELGraph):
            raise TypeError("{} is not a {}".format(other, self.__class__.__name__))

        left_outer_join(self, other)
        return self

    def __xor__(self, other: "BELGraph") -> "BELGraph":
        """Join this graph with another using :func:`pybel.struct.left_node_intersection_join`.

        :param other: Another BEL graph

        Example usage:

        >>> from pybel.examples import ras_tloc_graph, braf_graph
        >>> k = ras_tloc_graph ^ braf_graph
        """
        if not isinstance(other, BELGraph):
            raise TypeError("{} is not a {}".format(other, self.__class__.__name__))

        return left_node_intersection_join(self, other)

    @staticmethod
    def node_to_bel(n: BaseEntity) -> str:
        """Serialize a node as BEL."""
        warnings.warn("use node.as_bel()", DeprecationWarning)
        return n.as_bel()

    @staticmethod
    def edge_to_bel(
        u: BaseEntity,
        v: BaseEntity,
        edge_data: EdgeData,
        sep: Optional[str] = None,
        use_identifiers: bool = True,
    ) -> str:
        """Serialize a pair of nodes and related edge data as a BEL relation."""
        return edge_to_bel(u, v, data=edge_data, sep=sep, use_identifiers=use_identifiers)

    def _has_no_equivalent_edge(self, u: BaseEntity, v: BaseEntity) -> bool:
        return not any(EQUIVALENT_TO == data[RELATION] for data in self[u][v].values())

    def _equivalent_node_iterator_helper(self, node: BaseEntity, visited: Set[BaseEntity]) -> BaseEntity:
        """Iterate over nodes and their data that are equal to the given node, starting with the original."""
        for v in self[node]:
            if v in visited:
                continue

            if self._has_no_equivalent_edge(node, v):
                continue

            visited.add(v)
            yield v
            yield from self._equivalent_node_iterator_helper(v, visited)

    def iter_equivalent_nodes(self, node: BaseEntity) -> Iterable[BaseEntity]:
        """Iterate over nodes that are equivalent to the given node, including the original."""
        yield node
        yield from self._equivalent_node_iterator_helper(node, {node})

    def get_equivalent_nodes(self, node: BaseEntity) -> Set[BaseEntity]:
        """Get a set of equivalent nodes to this node, excluding the given node."""
        if isinstance(node, BaseEntity):
            return set(self.iter_equivalent_nodes(node))

        return set(self.iter_equivalent_nodes(node))

    @staticmethod
    def _node_has_namespace_helper(node: BaseEntity, namespace: str) -> bool:
        """Check that the node has namespace information.

        Might have cross references in future.
        """
        return isinstance(node, BaseConcept) and node.namespace.lower() == namespace.lower()

    def node_has_namespace(self, node: BaseEntity, namespace: str) -> bool:
        """Check if the node have the given namespace.

        This also should look in the equivalent nodes.
        """
        return any(self._node_has_namespace_helper(n, namespace) for n in self.iter_equivalent_nodes(node))

    def _describe_list(self) -> List[Tuple[str, float]]:
        """Return useful information about the graph as a list of tuples."""
        warnings.warn("use graph.summary.list()", DeprecationWarning)
        return self.summarize.list()

    def summary_dict(self) -> Mapping[str, float]:
        """Return a dictionary that summarizes the graph."""
        warnings.warn("use graph.summary.dict()", DeprecationWarning)
        return self.summarize.dict()

    def summary_str(self) -> str:
        """Return a string that summarizes the graph."""
        warnings.warn("use graph.summary.str()", DeprecationWarning)
        return self.summarize.str()

    def ground(self, **kwargs) -> "BELGraph":
        """Ground this graph."""
        try:
            from ..grounding import ground
        except ImportError:
            logger.warning("Must install pyobo and protmapper, use pip install pybel[grounding] extra.")
            raise
        return ground(self, **kwargs)

    def _clean_annotations(self, annotations_dict: AnnotationsHint) -> AnnotationsDict:
        """Fix the formatting of annotation dict.

        .. seealso:: https://github.com/vtoure/bep/blob/master/docs/published/BEP-0013.md

        Scenarios:

        1. ``DEFINE ANNOTATION CellLine AS URL "..."``
           ``SET CellLine = "NIH-3T3 cell"``
           ``{'CellLine': dict(namespace='CellLine', identifier=None, name='NIH-3T3 cell')}``
        2. ``DEFINE ANNOTATION CellLine AS CURIE``
           ``SET CellLine = "bto:0000944 ! NIH-3T3 cell"``
           ``{'CellLine': dict(namespace='bto', identifier='0000944', name='NIH-3T3 cell')}``
        3. ``DEFINE ANNOTATION ECO AS MIRIAM``
           ``SET ECO = "0007682 ! reporter gene assay evidence used in manual assertion"``
           ``{'ECO': dict(namespace='ECO', identifier='0007682', name='reporter gene assay...')}``
        """
        return {
            key: sorted(
                self._clean_value(key, values),
                key=lambda e: (e.namespace, e.identifier, e.name),
            )
            for key, values in annotations_dict.items()
        }

    def _clean_value(
        self,
        key,
        values: Union[str, Entity, List[str], List[Mapping[str, str]], List[Entity]],
    ) -> List[Entity]:
        if key in self.annotation_miriam:  # this annotation was given by a lookup
            return self._clean_value_helper(key=key, namespace=key, values=values)
        if key in self.annotation_curie:
            if isinstance(values, Entity):
                return [values]
            if all(isinstance(v, dict) for v in values):
                return [Entity(**v) for v in values]
            if not all(isinstance(v, Entity) for v in values):
                raise ValueError("if annotation_curie, all must be given as Entity instances")
            return values
        if key in self.annotation_miriam:
            raise NotImplementedError("parsing of annotation as MIRIAM not yet implemented")
        if key in self.annotation_list or key in self.annotation_pattern:
            if isinstance(values, str):
                return [Entity(namespace=key, identifier=values)]
            if isinstance(values, Entity):
                return [values]
            if all(isinstance(v, Entity) for v in values):
                return values
            if all(isinstance(v, dict) for v in values):
                return [Entity(**v) for v in values]
            if all(isinstance(v, str) for v in values):
                return [Entity(namespace=key, identifier=v) for v in values]
            raise TypeError(f"Mixed values: {values}")
        if key in self.annotation_url:  # this is a name given
            return self._clean_value_helper(key=key, namespace=key, values=values)

        if self.raise_on_missing_annotations:
            raise NotImplementedError(f"where is key {key}?")

        return self._clean_value_helper(key=key, namespace=key, values=values)

    @staticmethod
    def _clean_value_helper(key, namespace, values):
        if isinstance(values, str):
            return [
                Entity(namespace=namespace, identifier=values),
            ]
        if isinstance(values, (list, set)):
            if all(isinstance(v, str) for v in values):
                return [Entity(namespace=namespace, identifier=identifier) for identifier in sorted(values)]
            elif all(isinstance(v, Entity) for v in values):
                return values
            elif all(isinstance(v, dict) for v in values):
                return [Entity(**v) for v in values]
            else:
                raise TypeError(f"list of wrong format for key {key}: {values}")
        if isinstance(values, dict):
            if all(isinstance(v, bool) for v in values.values()):
                return [Entity(namespace=namespace, identifier=identifier) for identifier in sorted(values)]
            raise TypeError(f"dictionary of wrong format for key {key}: {values}")
        raise TypeError(f"values of wrong data type for key {key}: {values}")


def _handle_modifier(side_data: Dict[str, Any]) -> Mapping[str, Any]:
    modifier = side_data.get(MODIFIER)
    effect = side_data.get(EFFECT)

    if modifier == ACTIVITY:
        if effect is not None:
            side_data[EFFECT] = Entity(**effect)
    elif modifier == TRANSLOCATION:
        if effect is not None:
            effect[FROM_LOC] = Entity(**effect[FROM_LOC])
            effect[TO_LOC] = Entity(**effect[TO_LOC])
    elif modifier == DEGRADATION or modifier is None:
        pass
    else:
        raise ValueError("invalid modifier: {}".format(modifier))

    if LOCATION in side_data:
        side_data[LOCATION] = Entity(**side_data[LOCATION])
    return side_data


def _handle_citation(citation: Union[str, Tuple[str, str], CitationDict]) -> CitationDict:
    if isinstance(citation, str):
        return citation_dict(namespace=CITATION_TYPE_PUBMED, identifier=citation)
    elif isinstance(citation, tuple):
        return citation_dict(namespace=citation[0], identifier=citation[1])
    elif isinstance(citation, CitationDict):
        return citation
    elif isinstance(citation, dict):
        return CitationDict(**citation)
    elif citation is None:
        raise ValueError("citation was None")
    else:
        raise TypeError(f"citation is the wrong type: {citation}")


class Dispatch:
    def __init__(self, graph: BELGraph):
        self.graph = graph


class CountDispatch(Dispatch):
    """A dispatch for count functions that can be found at :data:`pybel.BELGraph.count`."""

    def functions(self) -> Counter:
        """Count the functions in a graph.

        >>> from pybel.examples import sialic_acid_graph
        >>> sialic_acid_graph.count.functions()
        Counter({'Protein': 7, 'Complex': 1, 'Abundance': 1})
        """
        from .summary import count_functions

        return count_functions(self.graph)

    def namespaces(self) -> Counter:
        """Return a counter of namespaces' occurrences in nodes in the graph."""
        from .summary import count_namespaces

        return count_namespaces(self.graph)

    def pathologies(self) -> Counter:
        """Return a counter of pathologies' occurrences in edges in the graph."""
        from .summary import count_pathologies

        return count_pathologies(self.graph)

    def annotations(self) -> Counter:
        """Return a counter of annotations' occurrences in edges in the graph."""
        from .summary import count_annotations

        return count_annotations(self.graph)

    def variants(self) -> Counter:
        """Return a counter of variants' occurrences in nodes in the graph."""
        from .summary import count_variants

        return count_variants(self.graph)

    def relations(self) -> Counter:
        """Return a counter of relations' occurrences in edges in the graph."""
        from .summary import count_relations

        return count_relations(self.graph)

    def error_types(self) -> Counter:
        """Return a counter of error types' occurrences in BEL script underlying the graph."""
        from .summary import count_error_types

        return count_error_types(self.graph)

    def names_by_namespace(self, namespace: str) -> Counter:
        from .summary import count_names_by_namespace

        return count_names_by_namespace(self.graph, namespace=namespace)

    def modifications(self) -> Counter:
        """Return a counter of relation modifications' occurrences (activity, translocation, etc.) in the graph."""
        from .summary.node_summary import count_modifications

        return count_modifications(self.graph)

    def authors(self) -> Counter:
        """Return a counter of the number of edges to which each author contributed in the graph."""
        return Counter(_iterate_authors(self.graph))

    def citations(self) -> int:
        """Return the number of citations."""
        return len(set(_iterate_citations(self.graph)))


def _iterate_citations(graph: BELGraph) -> Iterable[Tuple[str, str]]:
    for _, _, data in graph.edges(data=True):
        if CITATION in data:
            yield data[CITATION][NAMESPACE], data[CITATION][IDENTIFIER]


def _iterate_authors(graph: BELGraph) -> Iterable[str]:
    return chain.from_iterable(
        data[CITATION][CITATION_AUTHORS]
        for _, _, data in graph.edges(data=True)
        if CITATION in data and CITATION_AUTHORS in data[CITATION]
    )


class SummarizeDispatch(Dispatch):
    """A dispatch for summary printing functions that can be found at :data:`pybel.BELGraph.summarize`."""

    def __call__(self, file: Optional[TextIO] = None, examples: bool = True) -> None:
        self.statistics(file=file)
        print("", file=file)
        self.nodes(file=file, examples=examples)
        print("", file=file)
        self.namespaces(file=file, examples=examples)
        print("", file=file)
        self.edges(file=file, examples=examples)
        print("", file=file)

    def _repr_html_(self) -> str:
        from .summary import supersummary as ss

        return dedent(
            f"""\
            <h2>Metadata</h2>
            {tabulate(self._metadata_list(), tablefmt='html')}
            <h2>Statistics</h2>
            {tabulate(self._statistics_list(prose_prefix=False), tablefmt='html')}
            <h2>Nodes</h2>
            {ss.functions_str(self.graph, examples=True, add_count=False, tablefmt='html')}
            <h2>Namespaces</h2>
            {ss.namespaces_str(self.graph, examples=True, add_count=False, tablefmt='html')}
            <h2>Edges</h2>
            {ss.edges_str(self.graph, examples=True, add_count=False, tablefmt='html')}
        """
        )

    def statistics(self, file: Optional[TextIO] = None):
        """Print summary statistics on the graph."""
        print(self.str(), file=file)

    def nodes(self, file: Optional[TextIO] = None, examples: bool = True):
        """Print a summary of the nodes' functions in the graph."""
        from .summary.supersummary import functions_str

        print(functions_str(self.graph, examples=examples), file=file)

    def namespaces(self, file: Optional[TextIO] = None, examples: bool = True):
        """Print a summary of the nodes' namespaces in the graph."""
        from .summary.supersummary import namespaces

        namespaces(self.graph, file=file, examples=examples)

    def edges(self, file: Optional[TextIO] = None, examples: bool = True):
        """Print a summary of the edges' types in the graph."""
        from .summary.supersummary import edges

        edges(self.graph, file=file, examples=examples)

    def citations(self, n: Optional[int] = 15, file: Optional[TextIO] = None):
        """Print a summary of the top citations' frequencies in the graph."""
        from .summary.supersummary import citations

        citations(self.graph, n=n, file=file)

    def dict(self) -> Mapping[str, float]:
        """Return a dictionary that summarizes the graph."""
        return dict(self.list())

    def str(self, **kwargs) -> str:
        """Return a string that summarizes the graph."""
        return tabulate(self.list(), **kwargs)

    def _metadata_list(self) -> List[Tuple[str, Any]]:
        rv = [
            ("Name", self.graph.name),
            ("Version", self.graph.version),
        ]
        if self.graph.authors:
            rv.append(("Authors", self.graph.authors))
        return rv

    def _statistics_list(self, prose_prefix: bool = True) -> List[Tuple[str, Any]]:
        number_nodes = self.graph.number_of_nodes()
        rv = [
            ("Nodes", number_nodes),
            ("Namespaces", len(self.graph.count.namespaces())),
            ("Edges", self.graph.number_of_edges()),
            ("Annotations", len(self.graph.count.annotations())),
            ("Citations", self.graph.number_of_citations()),
            ("Authors", self.graph.number_of_authors()),
            ("Components", nx.number_weakly_connected_components(self.graph)),
            ("Warnings", self.graph.number_of_warnings()),
        ]
        if prose_prefix:
            rv = [(f"Number of {x}", y) for x, y in rv]
        rv.append(("Network Density", "{:.2E}".format(nx.density(self.graph))))
        return rv

    def list(self) -> List[Tuple[str, Any]]:
        """Return a list of tuples that summarize the graph."""
        return [
            *self._metadata_list(),
            *self._statistics_list(),
        ]


class PlotDispatch(Dispatch):
    """A dispatch for count functions that can be found at :data:`pybel.BELGraph.plot`."""

    def summary(self, save: Optional[str] = None, **kwargs):
        """Plot a summary of the graph's nodes and edges using :mod:`matplotlib`."""
        from pybel_tools.summary.visualization import plot_summary

        fig, axes = plot_summary(self.graph, **kwargs)
        if save:
            fig.save(save)


class ExpandDispatch(Dispatch):
    """A dispatch for count functions that can be found at :data:`pybel.BELGraph.expand`."""

    @property
    def parent(self) -> BELGraph:
        """Get the parent BEL graph."""
        if not self.graph.parent:
            raise RuntimeError("Can not use expand dispatch on graph without a parent")
        return self.graph.parent

    def neighborhood(self, node: BaseEntity) -> BELGraph:
        """Expand around the neighborhood of a given node.

        >>> from pybel.examples import braf_graph
        >>> from pybel.dsl import Protein
        >>> thpo = Protein(namespace='HGNC', name='THPO', identifier='11795')
        >>> braf = Protein(namespace='HGNC', name='BRAF', identifier='1097')
        >>> raf1 = Protein(namespace='HGNC', name='RAF1', identifier='9829')
        >>> elk1 = Protein(namespace='HGNC', name='ELK1', identifier='3321')
        >>> subgraph_1 = braf_graph.induce.paths([braf, elk1])
        >>> assert thpo not in subgraph_1 and raf1 not in subgraph_1
        >>> subgraph_2 = subgraph_1.expand.neighborhood(braf)
        >>> assert thpo in subgraph_2 and raf1 not in subgraph_2
        """
        from .mutation import expand_node_neighborhood

        cp = self.graph.copy()
        expand_node_neighborhood(universe=self.parent, graph=cp, node=node)
        return cp

    def periphery(self, **kwargs):
        """Expand around the periphery of the graph w.r.t. its parent graph."""
        from pybel_tools.mutation.expansion import expand_periphery

        cp = self.graph.copy()
        expand_periphery(universe=self.parent, graph=cp, **kwargs)
        return cp

    def internal(self, **kwargs):
        """Expand missing edges between nodes in the graph w.r.t. its parent graph."""
        from pybel_tools.mutation.expansion import expand_internal

        cp = self.graph.copy()
        expand_internal(universe=self.parent, graph=cp, **kwargs)
        return cp


class InduceDispatch(Dispatch):
    """A dispatch for induction functions that can be found at :data:`pybel.BELGraph.induce`."""

    def paths(self, nodes: Iterable[BaseEntity]) -> Optional[BELGraph]:
        """Induce a subgraph on shortest paths between the nodes."""
        from .mutation import get_subgraph_by_all_shortest_paths

        return get_subgraph_by_all_shortest_paths(self.graph, nodes)

    def neighborhood(self, nodes: Iterable[BaseEntity]) -> Optional[BELGraph]:
        """Induce a subgraph around the neighborhood."""
        from .mutation import get_subgraph_by_neighborhood

        return get_subgraph_by_neighborhood(self.graph, nodes)

    def random(self, **kwargs) -> Optional[BELGraph]:
        """Induce a random subgraph."""
        from .mutation import get_random_subgraph

        return get_random_subgraph(self.graph, **kwargs)

    def annotation(self, prefix: str, identifier: str) -> Optional[BELGraph]:
        """Induce a subgraph on edges with the given annotation."""
        from .mutation import get_subgraph_by_annotation_value

        return get_subgraph_by_annotation_value(self.graph, prefix, identifier)



# -*- coding: utf-8 -*-

"""Tests for summary functions for nodes."""

import unittest
from collections import Counter

from pybel import BELGraph
from pybel.constants import ABUNDANCE, BIOPROCESS, COMPLEX, PROTEIN
from pybel.dsl import fusion_range, pathology, protein, protein_fusion
from pybel.examples import egf_graph, sialic_acid_graph
from pybel.struct.summary.node_summary import (
    count_functions,
    count_names_by_namespace,
    count_namespaces,
    count_pathologies,
    count_variants,
    get_functions,
    get_names,
    get_names_by_namespace,
    get_namespaces,
    get_top_hubs,
    get_top_pathologies,
)
from pybel.testing.utils import n


class TestSummary(unittest.TestCase):
    """Test node summary functions."""

    def test_functions_sialic(self):
        """Test counting nodes and grouping by function on the sialic acid graph."""
        result = {
            PROTEIN: 7,
            COMPLEX: 1,
            ABUNDANCE: 1,
        }

        self.assertEqual(set(result), get_functions(sialic_acid_graph))
        self.assertEqual(Counter(result), count_functions(sialic_acid_graph))

    def test_functions_egf(self):
        """Test counting nodes and grouping by function on the EGF graph."""
        result = {
            PROTEIN: 10,
            COMPLEX: 1,
            BIOPROCESS: 1,
        }

        self.assertEqual(set(result), get_functions(egf_graph))
        self.assertEqual(result, count_functions(egf_graph))

    def test_summarize_sialic(self):
        """Test getting and counting namespaces' contents on the sialic acid graph."""
        namespace_result = {
            "hgnc": 8,
            "chebi": 2,
            "go": 15,
        }
        self.assertEqual(set(namespace_result), get_namespaces(sialic_acid_graph))
        self.assertEqual(Counter(namespace_result), count_namespaces(sialic_acid_graph))

        hgnc_result = {
            "CD33": 3,  # once as reference, once in complex, and once as variant
            "TYROBP": 1,
            "SYK": 1,
            "PTPN6": 1,
            "PTPN11": 1,
            "TREM2": 1,
        }
        chebi_result = {
            "sialic acid": 2,
        }
        names = get_names(sialic_acid_graph)
        self.assertEqual(set(namespace_result), set(names))
        self.assertEqual(set(hgnc_result), names["hgnc"])
        self.assertEqual(set(chebi_result), names["chebi"])
        self.assertEqual(set(hgnc_result), get_names_by_namespace(sialic_acid_graph, "hgnc"))
        self.assertEqual(set(chebi_result), get_names_by_namespace(sialic_acid_graph, "chebi"))
        self.assertEqual(hgnc_result, dict(count_names_by_namespace(sialic_acid_graph, "hgnc")))
        self.assertEqual(chebi_result, dict(count_names_by_namespace(sialic_acid_graph, "chebi")))

    def test_namespaces_egf(self):
        """Test getting and counting namespaces' contents on the EGF graph."""
        result = {
            "hgnc": 15,
            "go": 7,
        }

        self.assertEqual(set(result), get_namespaces(egf_graph))
        self.assertEqual(Counter(result), count_namespaces(egf_graph))

    def test_names_fusions(self):
        """Test that names inside fusions are still found by the iterator."""
        graph = BELGraph()
        graph.namespace_url["HGNC"] = "http://dummy"

        node = protein_fusion(
            partner_5p=protein(name="A", namespace="HGNC"),
            range_5p=fusion_range("p", 1, 15),
            partner_3p=protein(name="B", namespace="HGNC"),
            range_3p=fusion_range("p", 1, 100),
        )

        graph.add_node_from_data(node)

        result = {
            "A": 1,
            "B": 1,
        }

        self.assertEqual(set(result), get_names_by_namespace(graph, "HGNC"))
        self.assertEqual(result, count_names_by_namespace(graph, "HGNC"))

    def test_get_names_raise(self):
        """Test that an index error is raised when trying to get names from a namespace that isn't present."""
        with self.assertRaises(IndexError):
            get_names_by_namespace(sialic_acid_graph, "NOPE")

    def test_count_names_raise(self):
        """Test that an index error is raised when trying to count a namespace that isn't present."""
        with self.assertRaises(IndexError):
            count_names_by_namespace(sialic_acid_graph, "NOPE")

    def test_count_variants(self):
        """Test counting the number of variants in a graph."""
        variants = count_variants(sialic_acid_graph)
        self.assertEqual(1, variants["pmod"])

    def test_count_pathologies(self):
        """Test counting pathologies in the graph."""
        graph = BELGraph()
        a, b = (protein(namespace="HGNC", name=n()) for _ in range(2))
        c, d = (pathology(namespace="DOID", name=n()) for _ in range(2))

        graph.add_association(a, c, citation=n(), evidence=n())
        graph.add_association(a, d, citation=n(), evidence=n())
        graph.add_association(b, d, citation=n(), evidence=n())

        pathology_counter = count_pathologies(graph)
        self.assertIn(c, pathology_counter)
        self.assertIn(d, pathology_counter)
        self.assertEqual(1, pathology_counter[c])
        self.assertEqual(2, pathology_counter[d])

        top_pathology_counter = get_top_pathologies(graph, n=1)
        self.assertEqual(1, len(top_pathology_counter))
        node, count = top_pathology_counter[0]
        self.assertEqual(d, node)
        self.assertEqual(2, count)

    def test_get_top_hubs(self):
        """Test counting pathologies in the graph."""
        graph = BELGraph()
        a, b, c = protein(n(), n()), protein(n(), n()), pathology(n(), n())

        graph.add_association(a, b, citation=n(), evidence=n())
        graph.add_association(a, c, citation=n(), evidence=n())

        top_hubs = get_top_hubs(graph, n=1)
        self.assertEqual(1, len(top_hubs))
        node, degree = top_hubs[0]
        self.assertEqual(a, node)
        self.assertEqual(4, degree)  # higher than expected because association edges



# -*- coding: utf-8 -*-

"""Warnings for old TSV conversion module."""

import warnings

from ..triples.converters import _safe_label

__all__ = [
    "_safe_label",
]

warnings.warn(
    """Use pybel.io.triples module instead. Changes in PyBEL v0.15.0:

  - pybel.io.tsv.converters._safe_label renamed to pybel.io.triples.converters._safe_label

Will be removed in PyBEL v0.16.*
  """,
    DeprecationWarning,
)



# -*- coding: utf-8 -*-

"""This module wraps conversion between :class:`pybel.BELGraph` and the Cyberinfrastructure Exchange (CX) JSON.

CX is an aspect-oriented network interchange format encoded in JSON with a format inspired by the JSON-LD encoding of
Resource Description Framework (RDF). It is primarily used by the Network Data Exchange (NDEx) and more recent versions
of Cytoscape.

.. seealso::

    - The NDEx Data Model `Specification <http://www.home.ndexbio.org/data-model/>`_
    - `Cytoscape.js <http://js.cytoscape.org/>`_
    - CX Support for Cytoscape.js on the Cytoscape `App Store <http://apps.cytoscape.org/apps/cxsupport>`_
"""

import gzip
import json
import logging
import time
from collections import defaultdict
from operator import methodcaller
from typing import Dict, List, Mapping, Optional, TextIO, Union

from networkx.utils import open_file

from ..canonicalize import calculate_canonical_name
from ..constants import (
    ANNOTATIONS,
    CITATION,
    EVIDENCE,
    FUSION,
    GRAPH_ANNOTATION_LIST,
    GRAPH_ANNOTATION_PATTERN,
    GRAPH_ANNOTATION_URL,
    GRAPH_METADATA,
    GRAPH_NAMESPACE_PATTERN,
    GRAPH_NAMESPACE_URL,
    MEMBERS,
    NAME,
    PARTNER_3P,
    PARTNER_5P,
    PRODUCTS,
    RANGE_3P,
    RANGE_5P,
    REACTANTS,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    UNQUALIFIED_EDGES,
    VARIANTS,
)
from ..dsl import BaseAbundance, BaseEntity
from ..language import Entity
from ..struct import BELGraph
from ..tokens import parse_result_to_dsl
from ..utils import expand_dict, flatten_dict

__all__ = [
    "to_cx",
    "to_cx_file",
    "to_cx_gz",
    "to_cx_jsons",
    "from_cx",
    "from_cx_file",
    "from_cx_gz",
    "from_cx_jsons",
    "NDEX_SOURCE_FORMAT",
]

log = logging.getLogger(__name__)

CX_NODE_NAME = "label"
NDEX_SOURCE_FORMAT = "ndex:sourceFormat"

NDEX_SOURCE_MODIFIER = "sourceModifier"
NDEX_TARGET_MODIFIER = "targetModifier"


def _cx_to_dict(list_of_dicts: List[Dict], key_tag: str = "k", value_tag: str = "v") -> Dict:
    """Convert a CX list of dictionaries to a flat dictionary."""
    return {d[key_tag]: d[value_tag] for d in list_of_dicts}


def _cleanse_fusion_dict(d: Dict) -> Dict:
    """Fix the fusion partner names."""
    return {k.replace("_", ""): v for k, v in d.items()}


_p_dict = {
    "partner5p": PARTNER_5P,
    "partner3p": PARTNER_3P,
    "range5p": RANGE_5P,
    "range3p": RANGE_3P,
}


def _restore_fusion_dict(d: Dict) -> Dict:
    return {_p_dict[k]: v for k, v in d.items()}


def build_node_mapping(graph: BELGraph) -> Mapping[BaseEntity, int]:
    """Build a mapping from a graph's nodes to their canonical sort order."""
    return {node: node_index for node_index, node in enumerate(sorted(graph, key=methodcaller("as_bel")))}


def to_cx(graph: BELGraph) -> List[Dict]:  # noqa: C901
    """Convert a BEL Graph to a CX JSON object for use with `NDEx <http://www.ndexbio.org/>`_.

    .. seealso::

        - `NDEx Python Client <https://github.com/ndexbio/ndex-python>`_
    """
    node_mapping = build_node_mapping(graph)
    node_index_data = {}
    nodes_entry = []
    node_attributes_entry = []

    for node, node_index in node_mapping.items():
        node_index_data[node_index] = node

        node_entry_dict = {
            "@id": node_index,
            "n": calculate_canonical_name(node),
        }

        if isinstance(node, BaseAbundance):
            node_entry_dict["r"] = node.curie

        nodes_entry.append(node_entry_dict)

        aliases = []
        if isinstance(node, BaseAbundance):
            aliases.extend(xref.curie for xref in node.xrefs)

        if aliases:
            node_attributes_entry.append(
                {
                    "po": node_index,
                    "n": "alias",
                    "v": aliases,
                    "d": "list_of_str",
                }
            )

        for k, v in node.items():
            if k == VARIANTS:
                for i, el in enumerate(v):
                    for a, b in flatten_dict(el).items():
                        node_attributes_entry.append(
                            {
                                "po": node_index,
                                "n": "{}_{}_{}".format(k, i, a),
                                "v": b,
                            }
                        )
            elif k == FUSION:
                v = _cleanse_fusion_dict(v)
                for a, b in flatten_dict(v).items():
                    node_attributes_entry.append(
                        {
                            "po": node_index,
                            "n": "{}_{}".format(k, a),
                            "v": b,
                        }
                    )

            elif k == NAME:
                node_attributes_entry.append(
                    {
                        "po": node_index,
                        "n": CX_NODE_NAME,
                        "v": v,
                    }
                )

            elif k in {PRODUCTS, REACTANTS, MEMBERS}:
                node_attributes_entry.append(
                    {
                        "po": node_index,
                        "n": k,
                        "v": json.dumps(v),
                    }
                )

            else:
                node_attributes_entry.append(
                    {
                        "po": node_index,
                        "n": k,
                        "v": v,
                    }
                )

    edges_entry = []
    edge_attributes_entry = []

    for edge_index, (source, target, d) in enumerate(graph.edges(data=True)):
        uid = node_mapping[source]
        vid = node_mapping[target]

        edges_entry.append(
            {
                "@id": edge_index,
                "s": uid,
                "t": vid,
                "i": d[RELATION],
            }
        )

        if EVIDENCE in d:
            edge_attributes_entry.append(
                {
                    "po": edge_index,
                    "n": EVIDENCE,
                    "v": d[EVIDENCE],
                }
            )

            for k, v in d[CITATION].items():
                edge_attributes_entry.append(
                    {
                        "po": edge_index,
                        "n": "{}_{}".format(CITATION, k),
                        "v": v,
                    }
                )

        if ANNOTATIONS in d:
            for annotation, values in d[ANNOTATIONS].items():
                edge_attributes_entry.append(
                    {
                        "po": edge_index,
                        "n": annotation,
                        "v": sorted(values, key=lambda e: (e.namespace, e.identifier, e.name)),
                        "d": "list_of_string",
                    }
                )

        if SOURCE_MODIFIER in d:
            for k, v in flatten_dict(d[SOURCE_MODIFIER]).items():
                edge_attributes_entry.append(
                    {
                        "po": edge_index,
                        "n": "{}_{}".format(NDEX_SOURCE_MODIFIER, k),
                        "v": v,
                    }
                )

        if TARGET_MODIFIER in d:
            for k, v in flatten_dict(d[TARGET_MODIFIER]).items():
                edge_attributes_entry.append(
                    {
                        "po": edge_index,
                        "n": "{}_{}".format(NDEX_TARGET_MODIFIER, k),
                        "v": v,
                    }
                )

    context_legend = {}

    for key in graph.namespace_url:
        context_legend[key] = GRAPH_NAMESPACE_URL

    for key in graph.namespace_pattern:
        context_legend[key] = GRAPH_NAMESPACE_PATTERN

    for key in graph.annotation_url:
        context_legend[key] = GRAPH_ANNOTATION_URL

    for key in graph.annotation_pattern:
        context_legend[key] = GRAPH_ANNOTATION_PATTERN

    for key in graph.annotation_list:
        context_legend[key] = GRAPH_ANNOTATION_LIST

    context_legend_entry = []
    for keyword, resource_type in context_legend.items():
        context_legend_entry.append(
            {
                "k": keyword,
                "v": resource_type,
            }
        )

    annotation_list_keys_lookup = {keyword: i for i, keyword in enumerate(sorted(graph.annotation_list))}
    annotation_lists_entry = []
    for keyword, values in graph.annotation_list.items():
        for v in values:
            annotation_lists_entry.append(
                {
                    "k": annotation_list_keys_lookup[keyword],
                    "v": v,
                }
            )

    context_entry_dict = {}
    context_entry_dict.update(graph.namespace_url)
    context_entry_dict.update(graph.namespace_pattern)
    context_entry_dict.update(graph.annotation_url)
    context_entry_dict.update(graph.annotation_pattern)
    context_entry_dict.update(annotation_list_keys_lookup)

    context_entry_dict.update(graph.namespace_url)
    context_entry = [context_entry_dict]

    network_attributes_entry = [
        {
            "n": NDEX_SOURCE_FORMAT,
            "v": "PyBEL",
        }
    ]
    for k, v in graph.document.items():
        network_attributes_entry.append(
            {
                "n": k,
                "v": v,
            }
        )

    # Coalesce to cx
    # cx = create_aspect.number_verification()
    cx = [{"numberVerification": [{"longNumber": 281474976710655}]}]

    cx_pairs = [
        ("@context", context_entry),
        ("context_legend", context_legend_entry),
        ("annotation_lists", annotation_lists_entry),
        ("networkAttributes", network_attributes_entry),
        ("nodes", nodes_entry),
        ("nodeAttributes", node_attributes_entry),
        ("edges", edges_entry),
        ("edgeAttributes", edge_attributes_entry),
    ]

    cx_metadata = []

    for key, aspect in cx_pairs:
        aspect_dict = {
            "name": key,
            "elementCount": len(aspect),
            "lastUpdate": time.time(),
            "consistencyGroup": 1,
            "properties": [],
            "version": "1.0",
        }

        if key in {"citations", "supports", "nodes", "edges"}:
            aspect_dict["idCounter"] = len(aspect)

        cx_metadata.append(aspect_dict)

    cx.append(
        {
            "metaData": cx_metadata,
        }
    )

    for key, aspect in cx_pairs:
        cx.append(
            {
                key: aspect,
            }
        )

    cx.append({"status": [{"error": "", "success": True}]})

    return cx


@open_file(1, mode="w")
def to_cx_file(graph: BELGraph, path: Union[str, TextIO], indent: Optional[int] = 2, **kwargs) -> None:
    """Write a BEL graph to a JSON file in CX format.

    :param graph: A BEL graph
    :param path: A writable file or file-like
    :param indent: How many spaces to use to pretty print. Change to None for no pretty printing

    The example below shows how to output a BEL graph as CX to an open file.

    .. code-block:: python

       from pybel.examples import sialic_acid_graph
       from pybel import to_cx_file
       with open('graph.bel.cx.json', 'w') as file:
           to_cx_file(sialic_acid_graph, file)

    The example below shows how to output a BEL graph as CX to a file at a given path.

    .. code-block:: python

        from pybel.examples import sialic_acid_graph
        from pybel import to_cx_file
        to_cx_file(sialic_acid_graph, 'graph.bel.cx.json')

    If you have a big graph, you might consider storing it as a gzipped JGIF file
    by using :func:`to_cx_gz`.
    """
    graph_cx_json_dict = to_cx(graph)
    json.dump(graph_cx_json_dict, path, ensure_ascii=False, indent=indent, **kwargs)


def to_cx_gz(graph, path: str, **kwargs) -> None:
    """Write a graph as CX JSON to a gzip file."""
    with gzip.open(path, "wt") as file:
        json.dump(to_cx(graph), file, ensure_ascii=False, **kwargs)


def to_cx_jsons(graph: BELGraph, **kwargs) -> str:
    """Dump this graph as a CX JSON object to a string."""
    return json.dumps(to_cx(graph), ensure_ascii=False, **kwargs)


def _iterate_list_of_dicts(list_of_dicts: List[Dict]):
    """Iterate over a list of dictionaries.

    :type list_of_dicts: list[dict[A,B]]
    :rtype: iter[tuple[A,B]]
    """
    for dictionary in list_of_dicts:
        for key, value in dictionary.items():
            yield key, value


def from_cx(cx: List[Dict]) -> BELGraph:  # noqa: C901
    """Rebuild a BELGraph from CX JSON output from PyBEL.

    :param cx: The CX JSON object for this graph
    """
    graph = BELGraph()

    context_legend_aspect = []
    annotation_lists_aspect = []
    context_entry = {}
    network_attributes_aspect = []
    nodes_aspect = []
    node_attributes_aspect = []
    edge_annotations_aspect = []
    edges_aspect = []
    meta_entries = defaultdict(list)

    for key, value in _iterate_list_of_dicts(cx):
        if key == "context_legend":
            context_legend_aspect.extend(value)

        elif key == "annotation_lists":
            annotation_lists_aspect.extend(value)

        elif key == "@context":
            for element in value:
                context_entry.update(element)

        elif key == "networkAttributes":
            network_attributes_aspect.extend(value)

        elif key == "nodes":
            nodes_aspect.extend(value)

        elif key == "nodeAttributes":
            node_attributes_aspect.extend(value)

        elif key == "edges":
            edges_aspect.extend(value)

        elif key == "edgeAttributes":
            edge_annotations_aspect.extend(value)

        else:
            meta_entries[key].extend(value)

    context_legend = _cx_to_dict(context_legend_aspect)

    annotation_lists = defaultdict(set)
    for data in annotation_lists_aspect:
        annotation_lists[data["k"]].add(data["v"])

    for keyword, entry in context_entry.items():
        if context_legend[keyword] == GRAPH_NAMESPACE_URL:
            graph.namespace_url[keyword] = entry
        elif context_legend[keyword] == GRAPH_NAMESPACE_PATTERN:
            graph.namespace_pattern[keyword] = entry
        elif context_legend[keyword] == GRAPH_ANNOTATION_URL:
            graph.annotation_url[keyword] = entry
        elif context_legend[keyword] == GRAPH_ANNOTATION_PATTERN:
            graph.annotation_pattern[keyword] = entry
        elif context_legend[keyword] == GRAPH_ANNOTATION_LIST:
            graph.annotation_list[keyword] = annotation_lists[entry]

    for data in network_attributes_aspect:
        if data["n"] == NDEX_SOURCE_FORMAT:
            continue
        graph.graph[GRAPH_METADATA][data["n"]] = data["v"]

    node_name = {}
    for data in nodes_aspect:
        node_name[data["@id"]] = data["n"]

    node_data = defaultdict(dict)
    for data in node_attributes_aspect:
        node_data[data["po"]][data["n"]] = data["v"]

    # put all normal data here
    node_data_pp = defaultdict(dict)

    # Group all fusion-related data here
    node_data_fusion = defaultdict(dict)

    # Group all variant-related data
    node_data_variants = defaultdict(lambda: defaultdict(dict))

    for nid, data in node_data.items():
        for key, value in data.items():
            if key.startswith(FUSION):
                node_data_fusion[nid][key] = value
            elif key.startswith(VARIANTS):
                _, i, vls = key.split("_", 2)
                node_data_variants[nid][i][vls] = value
            elif key in {PRODUCTS, REACTANTS, MEMBERS}:
                node_data_pp[nid][key] = json.loads(value)
            else:
                node_data_pp[nid][key] = value

    for nid, data in node_data_fusion.items():
        data = expand_dict(data)
        data[FUSION] = _restore_fusion_dict(data[FUSION])
        node_data_pp[nid].update(data)

    for nid, data in node_data_variants.items():
        node_data_pp[nid][VARIANTS] = [expand_dict(value) for _, value in sorted(data.items())]

    nid_node_tuple = {}
    for nid, data in node_data_pp.items():
        if CX_NODE_NAME in data:
            data[NAME] = data.pop(CX_NODE_NAME)

        nid_node_tuple[nid] = _node = parse_result_to_dsl(data)
        graph.add_node_from_data(_node)

    edge_relation = {}
    eid_source_nid = {}
    eid_target_nid = {}
    for data in edges_aspect:
        eid = data["@id"]
        edge_relation[eid] = data["i"]
        eid_source_nid[eid] = data["s"]
        eid_target_nid[eid] = data["t"]

    edge_data = defaultdict(dict)  # type: Dict[str, Dict[str, str]]
    for data in edge_annotations_aspect:
        edge_data[data["po"]][data["n"]] = data["v"]

    edge_citation = defaultdict(dict)  # type: Dict[str, Dict[str, str]]
    edge_subject = defaultdict(dict)
    edge_object = defaultdict(dict)
    edge_annotations = defaultdict(lambda: defaultdict(dict))
    edge_data_pp = defaultdict(dict)

    for eid, data in edge_data.items():
        for key, value in data.items():
            if key.startswith(CITATION):
                vl = _after_underscore(key)
                edge_citation[eid][vl] = value
            elif key.startswith(NDEX_SOURCE_MODIFIER):
                vl = _after_underscore(key)
                edge_subject[eid][vl] = value
            elif key.startswith(NDEX_TARGET_MODIFIER):
                vl = _after_underscore(key)
                edge_object[eid][vl] = value
            elif key == EVIDENCE:
                edge_data_pp[eid][EVIDENCE] = value
            else:
                edge_annotations[eid][key] = value

    for eid, data in edge_citation.items():
        edge_data_pp[eid][CITATION] = data

    for eid, data in edge_subject.items():
        edge_data_pp[eid][SOURCE_MODIFIER] = expand_dict(data)

    for eid, data in edge_object.items():
        edge_data_pp[eid][TARGET_MODIFIER] = expand_dict(data)

    for eid in edge_relation:
        if eid in edge_annotations:  # FIXME stick this in edge_data.items() iteration
            edge_data_pp[eid][ANNOTATIONS] = {
                key: [Entity(**v) for v in values] for key, values in edge_annotations[eid].items()
            }

        if eid in edge_citation:
            graph.add_qualified_edge(
                nid_node_tuple[eid_source_nid[eid]],
                nid_node_tuple[eid_target_nid[eid]],
                relation=edge_relation[eid],
                citation=edge_data_pp[eid][CITATION],
                evidence=edge_data_pp[eid][EVIDENCE],
                source_modifier=edge_data_pp[eid].get(SOURCE_MODIFIER),
                target_modifier=edge_data_pp[eid].get(TARGET_MODIFIER),
                annotations=edge_data_pp[eid].get(ANNOTATIONS),
            )
        elif edge_relation[eid] in UNQUALIFIED_EDGES:
            graph.add_unqualified_edge(
                nid_node_tuple[eid_source_nid[eid]],
                nid_node_tuple[eid_target_nid[eid]],
                edge_relation[eid],
            )
        else:
            raise ValueError("problem adding edge: {}".format(eid))

    return graph


def _after_underscore(key):
    _, vl = key.split("_", 1)
    return vl


@open_file(0, mode="r")
def from_cx_file(path: Union[str, TextIO]) -> BELGraph:
    """Read a file containing CX JSON and converts to a BEL graph.

    :param path: A readable file or file-like containing the CX JSON for this graph
    :return: A BEL Graph representing the CX graph contained in the file
    """
    return from_cx(json.load(path))


def from_cx_gz(path: str) -> BELGraph:
    """Read a graph as CX JSON from a gzip file."""
    with gzip.open(path, "rt") as file:
        return from_cx(json.load(file))


def from_cx_jsons(graph_json_str: str) -> BELGraph:
    """Read a BEL graph from a CX JSON string."""
    return from_cx(json.loads(graph_json_str))



# -*- coding: utf-8 -*-

"""This module contains the functions for decorating transformation functions.

A transformation function takes in a :class:`pybel.BELGraph` and either returns None (in-place) or a new
:class:`pybel.BELGraph` (out-of-place).
"""

import logging
from inspect import signature

from .exc import MissingPipelineFunctionError, PipelineNameError

__all__ = [
    "in_place_transformation",
    "uni_in_place_transformation",
    "uni_transformation",
    "transformation",
    "get_transformation",
    "mapped",
    "has_arguments_map",
    "no_arguments_map",
]

logger = logging.getLogger(__name__)

mapped = {}
universe_map = {}
in_place_map = {}
has_arguments_map = {}
no_arguments_map = {}


def _has_arguments(func, universe):
    sig = signature(func)
    return (universe and 3 <= len(sig.parameters)) or (not universe and 2 <= len(sig.parameters))


def _register_function(name: str, func, universe: bool, in_place: bool):
    """Register a transformation function under the given name.

    :param name: Name to register the function under
    :param func: A function
    :param universe:
    :param in_place:
    :return: The same function, with additional properties added
    """
    if name in mapped:
        mapped_func = mapped[name]
        raise PipelineNameError(
            "{name} is already registered with {func_mod}.{func_name}".format(
                name=name,
                func_mod=mapped_func.__module__,
                func_name=mapped_func.__name__,
            ),
        )

    mapped[name] = func

    if universe:
        universe_map[name] = func

    if in_place:
        in_place_map[name] = func

    if _has_arguments(func, universe):
        has_arguments_map[name] = func
    else:
        no_arguments_map[name] = func

    return func


def _build_register_function(universe: bool, in_place: bool):  # noqa: D202
    """Build a decorator function to tag transformation functions.

    :param universe: Does the first positional argument of this function correspond to a universe graph?
    :param in_place: Does this function return a new graph, or just modify it in-place?
    """

    def register(func):
        """Tag a transformation function.

        :param func: A function
        :return: The same function, with additional properties added
        """
        return _register_function(func.__name__, func, universe, in_place)

    return register


#: A decorator for functions that modify BEL graphs in-place
in_place_transformation = _build_register_function(universe=False, in_place=True)
#: A decorator for functions that require a "universe" graph and modify BEL graphs in-place
uni_in_place_transformation = _build_register_function(universe=True, in_place=True)
#: A decorator for functions that require a "universe" graph and create new BEL graphs from old BEL graphs
uni_transformation = _build_register_function(universe=True, in_place=False)
#: A decorator for functions that create new BEL graphs from old BEL graphs
transformation = _build_register_function(universe=False, in_place=False)


def get_transformation(name: str):
    """Get a transformation function and error if its name is not registered.

    :param name: The name of a function to look up
    :return: A transformation function
    :raises MissingPipelineFunctionError: If the given function name is not registered
    """
    func = mapped.get(name)

    if func is None:
        raise MissingPipelineFunctionError("{} is not registered as a pipeline function".format(name))

    return func



# -*- coding: utf-8 -*-

"""This directory contains example networks, precompiled as BEL graphs that are appropriate to use in examples."""

from .ampk_example import ampk_graph
from .braf_example import braf_graph
from .egf_example import egf_graph
from .homology_example import homology_graph
from .sialic_acid_example import sialic_acid_graph
from .statin_example import statin_graph
from .tloc_example import ras_tloc_graph
from .vegf_example import vegf_graph

__all__ = [
    "egf_graph",
    "sialic_acid_graph",
    "statin_graph",
    "braf_graph",
    "homology_graph",
    "ras_tloc_graph",
    "ampk_graph",
    "vegf_graph",
]



# -*- coding: utf-8 -*-

"""An example describing a translocation.

.. code-block:: none

    SET Citation = {"PubMed", "16170185"}
    SET Evidence = "These modifications render Ras functional and capable of localizing to the lipid-rich inner surface of the cell membrane. The first and most critical modification, farnesylation, which is principally catalyzed by protein FTase, adds a 15-carbon hydrobobic farnesyl isoprenyl tail to the carboxyl terminus of Ras."
    SET TextLocation = Review

    cat(complex(p(HGNC:FNTA),p(HGNC:FNTB))) directlyIncreases p(SFAM:"RAS Family",pmod(F))
    p(SFAM:"RAS Family",pmod(F)) directlyIncreases tloc(p(SFAM:"RAS Family"),MESHCS:"Intracellular Space",MESHCS:"Cell Membrane")
"""

from ..dsl import (
    ComplexAbundance,
    Protein,
    ProteinModification,
    activity,
    translocation,
)
from ..language import Entity
from ..resources import FPLX_URL, GO_URL, HGNC_URL
from ..struct.graph import BELGraph

__all__ = [
    "ras_tloc_graph",
]

ras_tloc_graph = BELGraph(
    name="RAS Transocation Graph",
    version="1.0.1",
    description="The farnesylation of RAS causes its translocation to the cell membrane.",
    authors="Charles Tapley Hoyt",
    contact="cthoyt@gmail.com",
)

ras_tloc_graph.namespace_url.update(
    {
        "HGNC": HGNC_URL,
        "GO": GO_URL,
        "FPLX": FPLX_URL,
    }
)

evidence = "These modifications render Ras functional and capable of localizing to the lipid-rich inner surface of the cell membrane. The first and most critical modification, farnesylation, which is principally catalyzed by protein FTase, adds a 15-carbon hydrobobic farnesyl isoprenyl tail to the carboxyl terminus of Ras."
pmid = "16170185"

fnta = Protein(namespace="HGNC", name="FNTA", identifier="3782")
fntb = Protein(namespace="HGNC", name="FNTA", identifier="3785")
fnt = ComplexAbundance(namespace="FPLX", name="FNT", identifier="RAS", members=[fnta, fntb])
ras = Protein(namespace="FPLX", name="RAS", identifier="RAS")
ras_farn = ras.with_variants(ProteinModification("Farn"))

ras_tloc_graph.add_directly_increases(
    fnt,
    ras_farn,
    evidence=evidence,
    citation=pmid,
    source_modifier=activity("cat"),
)

ras_tloc_graph.add_directly_increases(
    ras_farn,
    ras,
    evidence=evidence,
    citation=pmid,
    target_modifier=translocation(
        from_loc=Entity(namespace="GO", name="intracellular", identifier="GO:0005622"),
        to_loc=Entity(namespace="GO", name="plasma membrane", identifier="GO:0005886"),
    ),
)



# -*- coding: utf-8 -*-

"""Warnings for old TSV conversion module."""

import warnings

from ..triples.api import to_triple as get_triple
from ..triples.api import to_triples as get_triples

__all__ = [
    "get_triple",
    "get_triples",
]

warnings.warn(
    """Use pybel.io.triples module instead. Changes in PyBEL v0.15.0:

  - pybel.io.tsv.api.get_triples renamed to pybel.to_triples
  - pybel.io.tsv.api.get_triple renamed to pybel.io.triples.to_triple

Will be removed in PyBEL v0.16.*
  """,
    DeprecationWarning,
)



# -*- coding: utf-8 -*-

"""Tests for input and output."""

import logging
import os
import re
import tempfile
import unittest
from io import BytesIO, StringIO
from pathlib import Path

import pybel
from pybel import (
    BELGraph,
    from_bel_script,
    from_bel_script_url,
    from_bytes,
    from_nodelink,
    from_nodelink_file,
    from_pickle,
    to_bel_script_lines,
    to_bytes,
    to_csv,
    to_graphml,
    to_gsea,
    to_nodelink,
    to_nodelink_file,
    to_pickle,
    to_sif,
)
from pybel.config import PYBEL_MINIMUM_IMPORT_VERSION
from pybel.constants import (
    ANNOTATIONS,
    CITATION,
    DECREASES,
    DIRECTLY_DECREASES,
    EVIDENCE,
    GRAPH_ANNOTATION_MIRIAM,
    GRAPH_PYBEL_VERSION,
    INCREASES,
    RELATION,
)
from pybel.dsl import BaseEntity, Gene, Protein
from pybel.examples.sialic_acid_example import sialic_acid_graph
from pybel.exceptions import (
    BELSyntaxError,
    InvalidFunctionSemantic,
    MissingCitationException,
    MissingNamespaceRegexWarning,
)
from pybel.io.exc import ImportVersionWarning, import_version_message_fmt
from pybel.io.line_utils import parse_lines
from pybel.io.nodelink import from_nodelink_jsons, to_nodelink_jsons
from pybel.language import Entity
from pybel.parser import BELParser
from pybel.struct.summary import get_syntax_errors
from pybel.testing.cases import TemporaryCacheClsMixin, TemporaryCacheMixin
from pybel.testing.constants import (
    test_bel_isolated,
    test_bel_misordered,
    test_bel_simple,
    test_bel_slushy,
    test_bel_thorough,
    test_bel_with_obo,
)
from pybel.testing.mocks import mock_bel_resources
from tests.constants import (
    BelReconstitutionMixin,
    TestTokenParserBase,
    akt1,
    casp8,
    citation_1,
    egfr,
    evidence_1,
    fadd,
    test_citation_dict,
    test_evidence_text,
    test_set_evidence,
)

logging.getLogger("requests").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

testan1 = "1"


class TestExampleInterchange(unittest.TestCase):
    """Test round-trip interchange of the sialic acid graph example."""

    def _help_test_equal(self, graph: BELGraph):
        """Check that a graph is equal to the sialic acid graph example."""
        for node in graph:
            self.assertIsInstance(node, BaseEntity)

        self.assertEqual(set(sialic_acid_graph), set(graph))
        self.assertEqual(set(sialic_acid_graph.edges()), set(graph.edges()))

        for node in sialic_acid_graph:
            if not isinstance(node, Protein):
                continue
            if node.namespace == "HGNC" and node.name == "CD33" and not node.variants:
                self.assertIsNotNone(node.xrefs)
                self.assertEqual(1, len(node.xrefs))

    def test_example_bytes(self):
        """Test the round-trip through bytes."""
        graph_bytes = to_bytes(sialic_acid_graph)
        graph = from_bytes(graph_bytes)
        self._help_test_equal(graph)

    def test_thorough_bytes_gz(self):
        """Test the round-trip through gzipped bytes."""
        graph_bytes = pybel.to_bytes_gz(sialic_acid_graph)
        graph = pybel.from_bytes_gz(graph_bytes)
        self._help_test_equal(graph)

    def test_example_pickle(self):
        """Test the round-trip through a pickle."""
        bio = BytesIO()
        to_pickle(sialic_acid_graph, bio)
        bio.seek(0)
        graph = from_pickle(bio)
        self._help_test_equal(graph)

    def test_example_pickle_gz(self):
        """Test the round-trip through a gzipped pickle."""
        with tempfile.TemporaryDirectory() as directory:
            path = os.path.join(directory, "test.gz")
            pybel.to_pickle_gz(sialic_acid_graph, path)
            graph = pybel.from_pickle_gz(path)
        self._help_test_equal(graph)

    def test_thorough_json(self):
        """Test the round-trip through node-link JSON."""
        graph_json_dict = to_nodelink(sialic_acid_graph)
        graph = from_nodelink(graph_json_dict)
        self._help_test_equal(graph)

    def test_thorough_jsons(self):
        """Test the round-trip through a node-link JSON string."""
        graph_json_str = to_nodelink_jsons(sialic_acid_graph)
        graph = from_nodelink_jsons(graph_json_str)
        self._help_test_equal(graph)

    def test_thorough_json_file(self):
        """Test the round-trip through a node-link JSON file."""
        sio = StringIO()
        to_nodelink_file(sialic_acid_graph, sio)
        sio.seek(0)
        graph = from_nodelink_file(sio)
        self._help_test_equal(graph)

    def test_thorough_sbel(self):
        """Test the round-trip through SBEL."""
        s = pybel.to_sbel(sialic_acid_graph)
        graph = pybel.from_sbel(s)
        self._help_test_equal(graph)

    def test_thorough_sbel_file(self):
        """Test the round-trip through a SBEL file."""
        sio = StringIO()
        pybel.to_sbel_file(sialic_acid_graph, sio)
        sio.seek(0)
        graph = pybel.from_sbel_file(sio)
        self._help_test_equal(graph)

    def test_thorough_sbel_gzip_path(self):
        """Test round trip through a SBEL gzipped file."""
        with tempfile.TemporaryDirectory() as directory:
            path = os.path.join(directory, "test.gzip")
            pybel.to_sbel_gz(sialic_acid_graph, path)
            graph = pybel.from_sbel_gz(path)
        self._help_test_equal(graph)


class TestInterchange(TemporaryCacheClsMixin, BelReconstitutionMixin):
    @classmethod
    def setUpClass(cls):
        """Set up this class with several pre-loaded BEL graphs."""
        super().setUpClass()

        with mock_bel_resources:
            cls.thorough_graph = from_bel_script(test_bel_thorough, manager=cls.manager, disallow_nested=False)
            cls.slushy_graph = from_bel_script(
                test_bel_slushy,
                manager=cls.manager,
                disallow_unqualified_translocations=True,
                disallow_nested=True,
            )
            cls.simple_graph = from_bel_script_url(Path(test_bel_simple).as_uri(), manager=cls.manager)
            cls.isolated_graph = from_bel_script(test_bel_isolated, manager=cls.manager)
            cls.misordered_graph = from_bel_script(test_bel_misordered, manager=cls.manager, citation_clearing=False)

    def test_thorough_path(self):
        self.bel_thorough_reconstituted(self.thorough_graph)

    def test_thorough_bytes(self):
        graph_bytes = to_bytes(self.thorough_graph)
        graph = from_bytes(graph_bytes)
        self.bel_thorough_reconstituted(graph)

    def test_thorough_pickle(self):
        bio = BytesIO()
        to_pickle(self.thorough_graph, bio)
        bio.seek(0)
        graph = from_pickle(bio)
        self.bel_thorough_reconstituted(graph)

    def test_thorough_json(self):
        graph_json_dict = to_nodelink(self.thorough_graph)
        graph = from_nodelink(graph_json_dict)
        self.bel_thorough_reconstituted(graph)

    def test_thorough_jsons(self):
        graph_json_str = to_nodelink_jsons(self.thorough_graph)
        graph = from_nodelink_jsons(graph_json_str)
        self.bel_thorough_reconstituted(graph)

    def test_thorough_json_file(self):
        sio = StringIO()
        to_nodelink_file(self.thorough_graph, sio)
        sio.seek(0)
        graph = from_nodelink_file(sio)
        self.bel_thorough_reconstituted(graph)

    def test_thorough_graphml(self):
        handle, path = tempfile.mkstemp()

        with open(path, "wb") as f:
            to_graphml(self.thorough_graph, f)

        os.close(handle)
        os.remove(path)

    def test_thorough_csv(self):
        handle, path = tempfile.mkstemp()

        with open(path, "w") as f:
            to_csv(self.thorough_graph, f)

        os.close(handle)
        os.remove(path)

    def test_thorough_sif(self):
        handle, path = tempfile.mkstemp()

        with open(path, "w") as f:
            to_sif(self.thorough_graph, f)

        os.close(handle)
        os.remove(path)

    def test_thorough_gsea(self):
        handle, path = tempfile.mkstemp()

        with open(path, "w") as f:
            to_gsea(self.thorough_graph, f)

        os.close(handle)
        os.remove(path)

    def test_thorough_upgrade(self):
        lines = to_bel_script_lines(self.thorough_graph, use_identifiers=True)
        reconstituted = BELGraph()
        parse_lines(reconstituted, lines, manager=self.manager)
        self.bel_thorough_reconstituted(reconstituted, check_citation_name=False, check_path=False)

    def test_slushy(self):
        self.bel_slushy_reconstituted(self.slushy_graph)

    def test_slushy_bytes(self):
        graph_bytes = to_bytes(self.slushy_graph)
        graph = from_bytes(graph_bytes)
        self.bel_slushy_reconstituted(graph)

    def test_slushy_syntax_errors(self):
        syntax_errors = get_syntax_errors(self.slushy_graph)
        for _, exc, _ in syntax_errors:
            self.assertIsInstance(exc, BELSyntaxError)
        self.assertEqual(1, len(syntax_errors))
        _, first_exc, _ = syntax_errors[0]
        self.assertEqual(98, first_exc.line_number)

    def test_slushy_json(self):
        graph_json = to_nodelink(self.slushy_graph)
        graph = from_nodelink(graph_json)
        self.bel_slushy_reconstituted(graph, check_warnings=False)

    def test_slushy_graphml(self):
        handle, path = tempfile.mkstemp()
        to_graphml(self.slushy_graph, path)
        os.close(handle)
        os.remove(path)

    def test_simple_compile(self):
        self.bel_simple_reconstituted(self.simple_graph)

    def test_isolated_compile(self):
        self.bel_isolated_reconstituted(self.isolated_graph)

    def test_isolated_upgrade(self):
        lines = to_bel_script_lines(self.isolated_graph)

        with mock_bel_resources:
            reconstituted = BELGraph()
            parse_lines(graph=reconstituted, lines=lines, manager=self.manager)

        self.bel_isolated_reconstituted(reconstituted)

    def test_misordered_compile(self):
        """Test that non-citation clearing mode works."""
        self.assertEqual(4, self.misordered_graph.number_of_nodes())
        self.assertEqual(3, self.misordered_graph.number_of_edges())

        e1 = {
            RELATION: INCREASES,
            CITATION: citation_1,
            EVIDENCE: evidence_1,
            ANNOTATIONS: {"TESTAN1": {testan1: True}},
        }
        self.assert_has_edge(self.misordered_graph, akt1, egfr, only=True, **e1)

        e2 = {
            RELATION: DECREASES,
            CITATION: citation_1,
            EVIDENCE: evidence_1,
            ANNOTATIONS: {"TESTAN1": {testan1: True}},
        }
        self.assert_has_edge(self.misordered_graph, egfr, fadd, only=True, **e2)

        e3 = {
            RELATION: DIRECTLY_DECREASES,
            CITATION: citation_1,
            EVIDENCE: evidence_1,
            ANNOTATIONS: {
                "TESTAN1": {testan1: True},
            },
        }
        self.assert_has_edge(self.misordered_graph, egfr, casp8, only=True, **e3)


namespace_to_term = {
    "TESTNS": {
        (None, "1"): "GRP",
        (None, "2"): "GRP",
    }
}

annotation_to_term = {
    "TestAnnotation1": {"A", "B", "C"},
    "TestAnnotation2": {"X", "Y", "Z"},
    "TestAnnotation3": {"D", "E", "F"},
}


class TestFull(TestTokenParserBase):
    @classmethod
    def setUpClass(cls):
        cls.parser = BELParser(
            graph=BELGraph(),  # gets overwritten in each test
            namespace_to_term_to_encoding=namespace_to_term,
            annotation_to_term=annotation_to_term,
            namespace_to_pattern={"dbSNP": re.compile("rs[0-9]*")},
        )

    def setUp(self):
        self.parser.clear()
        self.parser.graph = BELGraph()
        self.graph.annotation_list.update(
            {
                "TestAnnotation1": {"A", "B", "C"},
                "TestAnnotation2": {"X", "Y", "Z"},
                "TestAnnotation3": {"D", "E", "F"},
            }
        )
        self.parser.graph = self.graph
        self.assertIn(
            GRAPH_ANNOTATION_MIRIAM,
            self.graph.graph,
            msg=f"Graph metadata: {self.graph.graph}",
        )

    def test_regex_match(self):
        line = "g(dbSNP:rs10234) -- g(dbSNP:rs10235)"
        self.add_default_provenance()
        self.parser.parseString(line)
        self.assertIn(Gene("dbSNP", "rs10234"), self.parser.graph)
        self.assertIn(Gene("dbSNP", "rs10235"), self.parser.graph)

    def test_regex_mismatch(self):
        statement = "g(dbSNP:10234) -- g(dbSNP:rr10235)"
        with self.assertRaises(MissingNamespaceRegexWarning):
            self.parser.parseString(statement)

    def test_semantic_failure(self):
        self.assertIsNotNone(self.parser.concept_parser.namespace_to_name_to_encoding)
        self.assertIn("TESTNS", self.parser.concept_parser.namespace_to_name_to_encoding)
        self.assertIn("1", self.parser.concept_parser.namespace_to_name_to_encoding["TESTNS"])
        self.assertIn("2", self.parser.concept_parser.namespace_to_name_to_encoding["TESTNS"])
        statement = "bp(TESTNS:1) -- p(TESTNS:2)"
        with self.assertRaises(InvalidFunctionSemantic):
            self.parser.parseString(statement)

    def test_missing_citation(self):
        statements = [
            test_set_evidence,
            'SET TestAnnotation1 = "A"',
            'SET TestAnnotation2 = "X"',
            "g(TESTNS:1) -> g(TESTNS:2)",
        ]

        with self.assertRaises(MissingCitationException):
            self.parser.parse_lines(statements)

    def test_annotations(self):
        self.add_default_provenance()

        statements = [
            'SET TestAnnotation1 = "A"',
            'SET TestAnnotation2 = "X"',
            "g(TESTNS:1) -> g(TESTNS:2)",
        ]

        self.parser.parse_lines(statements)

        self.assertEqual(2, len(self.parser.control_parser.annotations))
        self.assertIn("TestAnnotation1", self.parser.control_parser.annotations)
        self.assertIn("TestAnnotation2", self.parser.control_parser.annotations)

        test_node_1 = Gene(namespace="TESTNS", name="1")
        test_node_2 = Gene(namespace="TESTNS", name="2")

        self.assertEqual(2, self.graph.number_of_nodes())
        self.assertIn(test_node_1, self.graph)
        self.assertIn(test_node_2, self.graph)

        self.assertEqual(1, self.parser.graph.number_of_edges())

        kwargs = {
            RELATION: INCREASES,
            ANNOTATIONS: {
                "TestAnnotation1": {"A": True},
                "TestAnnotation2": {"X": True},
            },
            EVIDENCE: test_evidence_text,
            CITATION: test_citation_dict,
        }

        self.assert_has_edge(test_node_1, test_node_2, only=True, **kwargs)

    def test_annotations_with_list(self):
        self.assertIsNotNone(self.parser.graph)
        self.add_default_provenance()

        statements = [
            'SET TestAnnotation1 = {"A","B"}',
            'SET TestAnnotation2 = "X"',
            "g(TESTNS:1) -> g(TESTNS:2)",
        ]

        self.parser.parse_lines(statements)

        self.assertEqual(2, len(self.parser.control_parser.annotations))
        self.assertIn("TestAnnotation1", self.parser.control_parser.annotations)
        self.assertIn("TestAnnotation2", self.parser.control_parser.annotations)
        self.assertEqual(2, len(self.parser.control_parser.annotations["TestAnnotation1"]))
        self.assertEqual(
            [
                Entity(namespace="TestAnnotation1", identifier="A"),
                Entity(namespace="TestAnnotation1", identifier="B"),
            ],
            self.parser.control_parser.annotations["TestAnnotation1"],
        )
        self.assertEqual(1, len(self.parser.control_parser.annotations["TestAnnotation2"]))
        self.assertEqual(
            [
                Entity(namespace="TestAnnotation2", identifier="X"),
            ],
            self.parser.control_parser.annotations["TestAnnotation2"],
        )

        test_node_1_dict = Gene(namespace="TESTNS", name="1")
        test_node_2_dict = Gene(namespace="TESTNS", name="2")

        self.assertEqual(2, self.parser.graph.number_of_nodes())
        self.assertIn(test_node_1_dict, self.graph)
        self.assertIn(test_node_2_dict, self.graph)

        self.assertEqual(1, self.parser.graph.number_of_edges())

        kwargs = {
            RELATION: INCREASES,
            EVIDENCE: test_evidence_text,
            ANNOTATIONS: {
                "TestAnnotation1": {"A": True, "B": True},
                "TestAnnotation2": {"X": True},
            },
            CITATION: test_citation_dict,
        }
        self.assert_has_edge(test_node_1_dict, test_node_2_dict, only=True, **kwargs)

    def test_annotations_with_multilist(self):
        self.add_default_provenance()

        statements = [
            'SET TestAnnotation1 = {"A","B"}',
            'SET TestAnnotation2 = "X"',
            'SET TestAnnotation3 = {"D","E"}',
            "g(TESTNS:1) -> g(TESTNS:2)",
        ]
        self.parser.parse_lines(statements)

        self.assertEqual(3, len(self.parser.control_parser.annotations))
        self.assertIn("TestAnnotation1", self.parser.control_parser.annotations)
        self.assertIn("TestAnnotation2", self.parser.control_parser.annotations)
        self.assertIn("TestAnnotation3", self.parser.control_parser.annotations)

        test_node_1 = Gene(namespace="TESTNS", name="1")
        test_node_2 = Gene(namespace="TESTNS", name="2")

        self.assertEqual(2, self.parser.graph.number_of_nodes())
        self.assertIn(test_node_1, self.graph)
        self.assertIn(test_node_2, self.graph)

        self.assertEqual(1, self.parser.graph.number_of_edges())

        kwargs = {
            RELATION: INCREASES,
            EVIDENCE: test_evidence_text,
            ANNOTATIONS: {
                "TestAnnotation1": {"A": True, "B": True},
                "TestAnnotation2": {"X": True},
                "TestAnnotation3": {"D": True, "E": True},
            },
            CITATION: test_citation_dict,
        }
        self.assert_has_edge(test_node_1, test_node_2, only=True, **kwargs)


class TestRandom(unittest.TestCase):
    def test_import_warning(self):
        """Tests an error is thrown when the version is set wrong"""
        graph = BELGraph()

        # Much with stuff that would normally be set
        graph.graph[GRAPH_PYBEL_VERSION] = "0.0.0"

        graph_bytes = to_bytes(graph)

        with self.assertRaises(ImportVersionWarning) as cm:
            from_bytes(graph_bytes)

            self.assertEqual(
                import_version_message_fmt.format("0.0.0", PYBEL_MINIMUM_IMPORT_VERSION),
                str(cm.exception),
            )


class TestNomenclature(TemporaryCacheMixin):
    """Test that `BEP-0008 nomenclature <http://bep.bel.bio/published/BEP-0008.html>`_ gets validated properly."""

    def test_bep_0008(self):
        """Test parsing works right"""
        graph = from_bel_script(test_bel_with_obo, manager=self.manager)
        self.assertIn("hgnc", graph.namespace_pattern)
        self.assertEqual(r"\d+", graph.namespace_pattern["hgnc"])

        self.assertEqual(0, graph.number_of_warnings(), msg=",\n".join(map(str, graph.warnings)))

        self.assertEqual(2, graph.number_of_nodes())
        self.assertIn(Protein(namespace="hgnc", identifier="391", name="AKT1"), graph)
        self.assertIn(Protein(namespace="hgnc", identifier="3236", name="EGFR"), graph)


if __name__ == "__main__":
    unittest.main()
